#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = ['LONG_PTR', 'UINT_PTR', 'LRESULT', 'GWLP_USERDATA', 'HCURSOR', 'PointerToID', 'WNDPROC', 'CALLWNDPROC', 'MARGINS', 'ACCENT_POLICY', 'WINCOMPATTRDATA', 'WINDOWPLACEMENT', 'PWINDOWPLACEMENT', 'LPWINDOWPLACEMENT', 'MONITORINFO', 'LPMONITORINFO', 'WNDCLASSEXW', 'PWNDCLASSEXW', 'WNDCLASS', 'CREATESTRUCTW', 'LPCREATESTRUCTW', 'WINDOWPOS', 'LPWINDOWPOS', 'PWINDOWPOS', 'NCCALCSIZE_PARAMS', 'LPNCCALCSIZE_PARAMS', 'STYLESTRUCT', 'LPSTYLESTRUCT', 'GUID']

import ctypes
from ctypes import wintypes, windll, WINFUNCTYPE
import win32gui, win32con, winerror
from enum import Enum, IntEnum
from time import sleep
from typing import Sequence, Dict, Optional
from cengal.data_generation.id_generator import IDGenerator

from ctypes import sizeof, byref, pointer, POINTER, c_int, c_long, c_longlong, c_ulonglong, c_uint, Structure, c_void_p
from ctypes.wintypes import ATOM, COLORREF, HANDLE, HBRUSH, HINSTANCE, HMENU, HWND, BOOL, HMONITOR, DWORD, WORD, BYTE, LPCVOID, LPCWSTR, LPMSG, LPRECT, LPVOID, LONG, MSG, RGB, UINT, WPARAM, LPARAM, HHOOK, ULONG, POINT, RECT, HICON, LPCSTR
from win32con import IDC_ARROW, CS_HREDRAW, CS_VREDRAW, GWL_STYLE, GWL_WNDPROC, SWP_FRAMECHANGED, SWP_NOMOVE, SWP_NOSIZE, SW_SHOW, SM_CXFRAME, SM_CYFRAME, GWL_USERDATA

from uuid import UUID


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


LONG_PTR = LPARAM
UINT_PTR = WPARAM
LRESULT = LONG_PTR

GWLP_USERDATA = GWL_USERDATA
# LONG_PTR = POINTER(wintypes.LPARAM)
# LONG_PTR = wintypes.LPARAM
# UINT_PTR = wintypes.WPARAM
# LRESULT = wintypes.LPARAM
# TRUE = win32con.TRUE
# FALSE = win32con.FALSE


# SW_SHOW = win32con.SW_SHOW


HCURSOR = HICON

PointerToID = POINTER(c_long)


# WNDPROC = ctypes.WINFUNCTYPE(c_int, c_long, c_int, c_int, c_int)
# WNDPROC = ctypes.WINFUNCTYPE(c_int, HWND, c_uint, wintypes.WPARAM, wintypes.LPARAM)
WNDPROC = ctypes.WINFUNCTYPE(LRESULT,  # return Value
                              HWND,     # First Param, the handle
                              UINT,     # second Param, message id
                              WPARAM,   # third param, additional message info (depends on message id)
                              LPARAM,   # fourth param, additional message info (depends on message id)
)
CALLWNDPROC = ctypes.WINFUNCTYPE(LRESULT, c_int, WPARAM, LPARAM)


# === STRUCTURES ===============================================================

# needs pointertomarginsstruct
class MARGINS(Structure):
    _fields_ = [("cxLeftWidth", c_int),
                ("cxRightWidth", c_int),
                ("cyTopHeight", c_int),
                ("cyBottomHeight", c_int)
                ]


# class ACCENTPOLICY(Structure):
#     _fields_ = [
#         ('AccentState', ULONG),
#         ('AccentFlags', ULONG),
#         ('GradientColor', ULONG),
#         ('AnimationId', ULONG),
#     ]
class ACCENT_POLICY(ctypes.Structure):
    _fields_ = [
        ("AccentState", wintypes.DWORD),
        ("AccentFlags", wintypes.DWORD),
        ("GradientColor", wintypes.DWORD),
        ("AnimationId", wintypes.DWORD),
        ("GradientAlpha", wintypes.DWORD),
        ("BlurPolicy", wintypes.DWORD),
    ]


class WINCOMPATTRDATA(Structure):
    _fields_ = [
        ('attribute', DWORD),
        ('pData', POINTER(ACCENT_POLICY)),
        ('dataSize', ULONG),
    ]


class WINDOWPLACEMENT(Structure):
    _fields_ = [
        ('length', UINT),
        ('flags', UINT),
        ('showCmd', UINT),
        ('ptMinPosition', POINT),
        ('ptMaxPosition', POINT),
        ('rcNormalPosition', RECT),
    ]

PWINDOWPLACEMENT = POINTER(WINDOWPLACEMENT)
LPWINDOWPLACEMENT = POINTER(WINDOWPLACEMENT)


class MONITORINFO(Structure):
    _fields_ = [
        ('cbSize', DWORD),
        ('rcMonitor', RECT),
        ('rcWork', RECT),
        ('dwFlags', DWORD),
    ]

LPMONITORINFO = POINTER(MONITORINFO)

class WNDCLASSEXW(Structure):
    _fields_ = [
        ('cbSize', UINT),
        ('style', UINT),
        ('lpfnWndProc', WNDPROC),
        ('cbClsExtra', c_int),
        ('cbWndExtra', c_int),
        ('hInstance', HANDLE),
        ('hIcon', HANDLE),
        ('hCursor', HANDLE),
        ('hbrBackground', wintypes.HBRUSH),
        ('lpszMenuName', wintypes.LPCWSTR),
        ('lpszClassName', wintypes.LPCWSTR),
        ('hIconSm', HICON),
    ]

PWNDCLASSEXW = POINTER(WNDCLASSEXW)

class WNDCLASS(Structure):
    _fields_ = [('style', c_uint),
            ('lpfnWndProc', WNDPROC),
            ('cbClsExtra', c_int),
            ('cbWndExtra', c_int),
            ('hInstance', HANDLE),
            ('hIcon', HANDLE),
            ('hCursor', HANDLE),
            ('hbrBackground', HANDLE),
            ('lpszMenuName', LPCSTR),
            ('lpszClassName', LPCSTR)]


# class CREATESTRUCTW(Structure):
#     _fields_ = [
#         ('lpCreateParams', c_void_p),
#         ('hInstance', wintypes.HINSTANCE),
#         ('hMenu', wintypes.HMENU),
#         ('hwndParent', HWND),
#         ('cy', c_int),
#         ('cx', c_int),
#         ('y', c_int),
#         ('x', c_int),
#         ('style', LONG),
#         ('lpszName', wintypes.LPCWSTR),
#         ('lpszClass', wintypes.LPCWSTR),
#         ('dwExStyle', DWORD),
#     ]

class CREATESTRUCTW(Structure):
    _fields_ = [
        ('lpCreateParams', c_void_p),
        ('hInstance', HANDLE),
        ('hMenu', HANDLE),
        ('hwndParent', HWND),
        ('cy', c_int),
        ('cx', c_int),
        ('y', c_int),
        ('x', c_int),
        ('style', c_int),
        ('lpszName', wintypes.LPCWSTR),
        ('lpszClass', wintypes.LPCWSTR),
        ('dwExStyle', c_int)
    ]

LPCREATESTRUCTW = POINTER(CREATESTRUCTW)


class WINDOWPOS(Structure):
    _fields_ = [
        ('hwnd', HWND),
        ('hwndInsertAfter', HWND),
        ('x', c_int),
        ('y', c_int),
        ('cx', c_int),
        ('cy', c_int),
        ('flags', UINT),
    ]

LPWINDOWPOS = POINTER(WINDOWPOS)
PWINDOWPOS = POINTER(WINDOWPOS)


class NCCALCSIZE_PARAMS(Structure):
    _fields_ = [
        ('rgrc', RECT * 3),
        ('lppos', PWINDOWPOS),
    ]

LPNCCALCSIZE_PARAMS = POINTER(NCCALCSIZE_PARAMS)


class STYLESTRUCT(Structure):
    _fields_ = [
        ('styleOld', DWORD),
        ('styleNew', DWORD),
    ]

LPSTYLESTRUCT = POINTER(STYLESTRUCT)


class GUID(Structure):
    _fields_ = [
        ("Data1", DWORD),
        ("Data2", WORD),
        ("Data3", WORD),
        ("Data4", BYTE * 8)
    ] 

    def __init__(self, uuid_):
        ctypes.Structure.__init__(self)
        self.Data1, self.Data2, self.Data3, self.Data4[0], self.Data4[1], rest = uuid_.fields
        for i in range(2, 8):
            self.Data4[i] = rest>>(8 - i - 1)*8 & 0xff

    @classmethod
    def from_uuid(cls, uuid_obj: UUID) -> 'GUID':
        return cls(uuid_obj)



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""


__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


__all__ = ['out_of_data_bounds', 'out_of_accessible_data_bounds', 'line_delimiter_n', 'line_delimiter_rn', 'line_delimiter', 'first_line', 'first_visible_line', 'line', 'last_line', 'last_visible_line', 'round_l', 'round_r', 'round', 'square_l', 'square_r', 'square', 'triangle_l', 'triangle_r', 'triangle', 'brace_l', 'brace_r', 'brace', 'quote', 'quotes', 'apostrophe', 'apostrophes', 'backtick', 'backticks']


from enum import Enum
from typing import Optional, Union
from .brackets import *


out_of_data_bounds: Bracket = Bracket(BracketAbsentType.out_of_data_bounds)
out_of_accessible_data_bounds: Bracket = Bracket(BracketAbsentType.out_of_accessible_data_bounds)
out_of_bounds: BracketsList = [out_of_accessible_data_bounds, out_of_data_bounds]

line_delimiter_n: Bracket = Bracket('\n')
line_delimiter_rn: Bracket = Bracket('\r\n')
line_delimiter: BracketsList = [line_delimiter_rn, line_delimiter_n]

line_brake: BracketsList = [line_delimiter_rn, line_delimiter_n, out_of_accessible_data_bounds, out_of_data_bounds]

first_line: BracketPair = BracketPair([out_of_data_bounds], line_delimiter)
first_visible_line: BracketPair = BracketPair([out_of_accessible_data_bounds], line_delimiter)
inner_line: BracketPair = BracketPair(line_delimiter, line_delimiter)
last_visible_line: BracketPair = BracketPair(line_delimiter, [out_of_accessible_data_bounds])
last_line: BracketPair = BracketPair(line_delimiter, [out_of_data_bounds])
line: BracketPair = BracketPair(line_brake, line_brake)

data_body: BracketPair = BracketPair([out_of_data_bounds], [out_of_data_bounds])
visible_body: BracketPair = BracketPair([out_of_accessible_data_bounds], [out_of_accessible_data_bounds])
body: BracketPair = BracketPair(out_of_bounds, out_of_bounds)

round_l: Bracket = Bracket('(')
round_r: Bracket = Bracket(')')
round: BracketPair = BracketPair([round_l], [round_r])

square_l: Bracket = Bracket('[')
square_r: Bracket = Bracket(']')
square: BracketPair = BracketPair([square_l], [square_r])

triangle_l: Bracket = Bracket('<')
triangle_r: Bracket = Bracket('>')
triangle: BracketPair = BracketPair([triangle_l], [triangle_r])

brace_l: Bracket = Bracket('{')
brace_r: Bracket = Bracket('}')
brace: BracketPair = BracketPair([brace_l], [brace_r])

quote: Bracket = Bracket('"')
quotes: BracketPair = BracketPair([quote], [quote])

apostrophe: Bracket = Bracket("'")
apostrophes: BracketPair = BracketPair([apostrophe], [apostrophe])

backtick: Bracket = Bracket('`')
backticks: BracketPair = BracketPair([backtick], [backtick])



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cengal.data_manipulation.serialization import *

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"

# from cengal.serialization import *
# s = Serializer(Serializers.msgpack)
# data = {1: 'Hello', 2: ('W', 0), 3: ['r', 1, {b'd': '.'}], 'To all!': b'!!1'}
# s.loads(s.dumps(data))


def _single_benchmark(title: str, desired_features: SerializerFeatures, test_data_type: TestDataType):
    print()
    print('~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~')
    print('~~~ {} ~~~'.format(title))
    best_serializer, all_results = get_most_efficient_serializers(
            desired_features, test_data_factory(test_data_type))
    output_template = 'Best match: {},\n\nAll results:\n{}'
    print(output_template.format(best_serializer, all_results))


if __name__ == '__main__':
    _single_benchmark('Binary, Small, Single Platform:',
                      {DataFormats.binary, Tags.current_platform},
                      TestDataType.small)
    _single_benchmark('Binary, Small, Multi Platform:',
                      {DataFormats.binary, Tags.multi_platform},
                      TestDataType.small)
    _single_benchmark('Text, Small, Multi Platform:',
                      {DataFormats.text, Tags.multi_platform},
                      TestDataType.small)

    print('\n\n============\n')

    _single_benchmark('Binary, Small, Deep, Single Platform:',
                      {DataFormats.binary, Tags.current_platform},
                      TestDataType.deep_small)
    _single_benchmark('Binary, Small, Deep, Multi Platform:',
                      {DataFormats.binary, Tags.multi_platform},
                      TestDataType.deep_small)
    _single_benchmark('Text, Small, Deep, Multi Platform:',
                      {DataFormats.text, Tags.multi_platform},
                      TestDataType.deep_small)

    print('\n\n============\n')

    _single_benchmark('Binary, Large, Single Platform:',
                      {DataFormats.binary, Tags.current_platform},
                      TestDataType.large)
    _single_benchmark('Binary, Large, Multi Platform:',
                      {DataFormats.binary, Tags.multi_platform},
                      TestDataType.large)
    _single_benchmark('Text, Large, Multi Platform:',
                      {DataFormats.text, Tags.multi_platform},
                      TestDataType.large)

    print('\n\n============\n')

    _single_benchmark('Binary, Large, Deep, Single Platform:',
                      {DataFormats.binary, Tags.current_platform},
                      TestDataType.deep_large)
    _single_benchmark('Binary, Large, Deep, Multi Platform:',
                      {DataFormats.binary, Tags.multi_platform},
                      TestDataType.deep_large)
    _single_benchmark('Text, Large, Deep, Multi Platform:',
                      {DataFormats.text, Tags.multi_platform},
                      TestDataType.deep_large)



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""


__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from cengal.unittest.behavior_stabilizer import UnittestTestCaseWithState, UnittestTestCaseState
from cengal.performance_test_lib import PrecisePerformanceTestTracer
from cengal.introspection.inspect import func_name, pdi, pifrl
from threading import Lock, RLock
from pprint import pprint

import difflib

from typing import Callable


def func_perf_test(func: Callable, *args, **kwargs):
    tr = PrecisePerformanceTestTracer(1.0)
    while tr.iter():
        func(*args, **kwargs)

    with tr as fast_iter:
        for i in fast_iter:
            func(*args, **kwargs)

    print('{}(): {} iter/s; {} seconds; {} iterations'.format(func_name(func), tr.iter_per_time_unit, tr.time_spent, tr.iterations_made))


from cengal.hardware.memory.shared_memory import *
from cengal.time_management.cpu_clock import cpu_clock
from cengal.time_management.high_precision_sync_sleep import hps_sleep

from dataclasses import dataclass
from typing import Any, Dict, Optional, Callable, Set, Type, Tuple, List, FrozenSet, Deque


class SimpleClass:
    def __init__(self, a, b):
        self.a = a
        self.b = b
    
    def __eq__(self, other):
        if isinstance(other, SimpleClass):
            return (self.a == other.a) and (self.b == other.b)
        
        return False


class SimpleClassWithSlots:
    __slots__ = ['a', 'b']

    def __init__(self, a, b):
        self.a = a
        self.b = b
    
    def __eq__(self, other):
        if isinstance(other, SimpleClass):
            return (self.a == other.a) and (self.b == other.b)
        
        return False


@dataclass
class DataClass:
    x: int
    y: 'Any'
    z: int = 5

    def add_one(self):
        self.x += 1


# Complex dataclass:

class Positions(IntEnum):
    manager = 0
    designer = 1
    accountant = 2
    lawyer = 3


class Employee:
    def __init__(self, name: str, age: int, position: Positions, years_of_employment: int = 0):
        self.name = name
        self.age = age
        self.position = position
        self.years_of_employment = years_of_employment
    
    def increase_years_of_employment(self):
        self.years_of_employment += 1
        self.age += 1
    
    def __repr__(self):
        return f'Employee(name={self.name}, age={self.age}, position={self.position}, years_of_employment={self.years_of_employment})'
    
    def __str__(self):
        return self.__repr__()


@dataclass
class CompanyInfo:
    company_id: int
    emails: Tuple[str, str]
    websites: List[str]
    income: float
    employees: int
    some_employee: Employee


@dataclass
class SomeSharedObject:
    some_processing_stage_control: bool
    str_value: str
    data_dict: Dict[Hashable, Any]
    company_info: CompanyInfo


_MAX_LENGTH = 80

DIFF_OMITTED = ('\nDiff is %s characters long. '
                 'Set self.maxDiff to None to see it.')


def safe_repr(obj, short=False):
    try:
        result = repr(obj)
    except Exception:
        result = object.__repr__(obj)
    if not short or len(result) < _MAX_LENGTH:
        return result
    return result[:_MAX_LENGTH] + ' [truncated]...'


class FakeTestCase:

    failureException = AssertionError

    longMessage = True

    _diffThreshold = 2**16

    maxDiff = 80*8

    def __init__(self):
        self._outcome = None
        self._testMethodDoc = 'No test'
        self._cleanups = []
        self._subtest = None

        # Map types to custom assertEqual functions that will compare
        # instances of said type in more detail to generate a more useful
        # error message.
        self._type_equality_funcs = {}
        self.addTypeEqualityFunc(dict, 'assertDictEqual')
        self.addTypeEqualityFunc(list, 'assertListEqual')
        self.addTypeEqualityFunc(tuple, 'assertTupleEqual')
        self.addTypeEqualityFunc(set, 'assertSetEqual')
        self.addTypeEqualityFunc(frozenset, 'assertSetEqual')
        self.addTypeEqualityFunc(str, 'assertMultiLineEqual')

    def _truncateMessage(self, message, diff):
        max_diff = self.maxDiff
        if max_diff is None or len(diff) <= max_diff:
            return message + diff
        return message + (DIFF_OMITTED % len(diff))

    def _baseAssertEqual(self, first, second, msg=None):
        """The default assertEqual implementation, not type specific."""
        if not first == second:
            standardMsg = str()
            msg = self._formatMessage(msg, standardMsg)
            raise self.failureException(msg)

    def assertMultiLineEqual(self, first, second, msg=None):
        """Assert that two multi-line strings are equal."""
        self.assertIsInstance(first, str, 'First argument is not a string')
        self.assertIsInstance(second, str, 'Second argument is not a string')

        if first != second:
            # don't use difflib if the strings are too long
            if (len(first) > self._diffThreshold or
                len(second) > self._diffThreshold):
                self._baseAssertEqual(first, second, msg)
            firstlines = first.splitlines(keepends=True)
            secondlines = second.splitlines(keepends=True)
            if len(firstlines) == 1 and first.strip('\r\n') == first:
                firstlines = [first + '\n']
                secondlines = [second + '\n']
            standardMsg = str()
            diff = '\n' + ''.join(difflib.ndiff(firstlines, secondlines))
            standardMsg = self._truncateMessage(standardMsg, diff)
            self.fail(self._formatMessage(msg, standardMsg))

    def assertSetEqual(self, set1, set2, msg=None):
        """A set-specific equality assertion.

        Args:
            set1: The first set to compare.
            set2: The second set to compare.
            msg: Optional message to use on failure instead of a list of
                    differences.

        assertSetEqual uses ducktyping to support different types of sets, and
        is optimized for sets specifically (parameters must support a
        difference method).
        """
        try:
            difference1 = set1.difference(set2)
        except TypeError as e:
            self.fail('invalid type when attempting set difference: %s' % e)
        except AttributeError as e:
            self.fail('first argument does not support set difference: %s' % e)

        try:
            difference2 = set2.difference(set1)
        except TypeError as e:
            self.fail('invalid type when attempting set difference: %s' % e)
        except AttributeError as e:
            self.fail('second argument does not support set difference: %s' % e)

        if not (difference1 or difference2):
            return

        lines = []
        if difference1:
            lines.append('Items in the first set but not the second:')
            for item in difference1:
                lines.append(repr(item))
        if difference2:
            lines.append('Items in the second set but not the first:')
            for item in difference2:
                lines.append(repr(item))

        standardMsg = '\n'.join(lines)
        self.fail(self._formatMessage(msg, standardMsg))

    def assertTupleEqual(self, tuple1, tuple2, msg=None):
        """A tuple-specific equality assertion.

        Args:
            tuple1: The first tuple to compare.
            tuple2: The second tuple to compare.
            msg: Optional message to use on failure instead of a list of
                    differences.
        """
        self.assertSequenceEqual(tuple1, tuple2, msg, seq_type=tuple)

    def assertSequenceEqual(self, seq1, seq2, msg=None, seq_type=None):
        """An equality assertion for ordered sequences (like lists and tuples).

        For the purposes of this function, a valid ordered sequence type is one
        which can be indexed, has a length, and has an equality operator.

        Args:
            seq1: The first sequence to compare.
            seq2: The second sequence to compare.
            seq_type: The expected datatype of the sequences, or None if no
                    datatype should be enforced.
            msg: Optional message to use on failure instead of a list of
                    differences.
        """
        if seq_type is not None:
            seq_type_name = seq_type.__name__
            if not isinstance(seq1, seq_type):
                raise self.failureException('First sequence is not a %s: %s'
                                        % (seq_type_name, safe_repr(seq1)))
            if not isinstance(seq2, seq_type):
                raise self.failureException('Second sequence is not a %s: %s'
                                        % (seq_type_name, safe_repr(seq2)))
        else:
            seq_type_name = "sequence"

        differing = None
        try:
            len1 = len(seq1)
        except (TypeError, NotImplementedError):
            differing = 'First %s has no length.    Non-sequence?' % (
                    seq_type_name)

        if differing is None:
            try:
                len2 = len(seq2)
            except (TypeError, NotImplementedError):
                differing = 'Second %s has no length.    Non-sequence?' % (
                        seq_type_name)

        if differing is None:
            if seq1 == seq2:
                return

            differing = str()

            for i in range(min(len1, len2)):
                try:
                    item1 = seq1[i]
                except (TypeError, IndexError, NotImplementedError):
                    differing += ('\nUnable to index element %d of first %s\n' %
                                 (i, seq_type_name))
                    break

                try:
                    item2 = seq2[i]
                except (TypeError, IndexError, NotImplementedError):
                    differing += ('\nUnable to index element %d of second %s\n' %
                                 (i, seq_type_name))
                    break

                if item1 != item2:
                    differing += str()
                    break
            else:
                if (len1 == len2 and seq_type is None and
                    type(seq1) != type(seq2)):
                    # The sequences are the same, but have differing types.
                    return

            if len1 > len2:
                differing += ('\nFirst %s contains %d additional '
                             'elements.\n' % (seq_type_name, len1 - len2))
                try:
                    differing += ('First extra element %d:\n%s\n' %
                                  (len2, safe_repr(seq1[len2])))
                except (TypeError, IndexError, NotImplementedError):
                    differing += ('Unable to index element %d '
                                  'of first %s\n' % (len2, seq_type_name))
            elif len1 < len2:
                differing += ('\nSecond %s contains %d additional '
                             'elements.\n' % (seq_type_name, len2 - len1))
                try:
                    differing += ('First extra element %d:\n%s\n' %
                                  (len1, safe_repr(seq2[len1])))
                except (TypeError, IndexError, NotImplementedError):
                    differing += ('Unable to index element %d '
                                  'of second %s\n' % (len1, seq_type_name))
        standardMsg = differing
        diffMsg = '\n' + '\n'.join(
            difflib.ndiff(pprint.pformat(seq1).splitlines(),
                          pprint.pformat(seq2).splitlines()))

        standardMsg = self._truncateMessage(standardMsg, diffMsg)
        msg = self._formatMessage(msg, standardMsg)
        self.fail(msg)

    def assertListEqual(self, list1, list2, msg=None):
        """A list-specific equality assertion.

        Args:
            list1: The first list to compare.
            list2: The second list to compare.
            msg: Optional message to use on failure instead of a list of
                    differences.

        """
        self.assertSequenceEqual(list1, list2, msg, seq_type=list)

    def addTypeEqualityFunc(self, typeobj, function):
        """Add a type specific assertEqual style function to compare a type.

        This method is for use by TestCase subclasses that need to register
        their own type equality functions to provide nicer error messages.

        Args:
            typeobj: The data type to call this function on when both values
                    are of the same type in assertEqual().
            function: The callable taking two arguments and an optional
                    msg= argument that raises self.failureException with a
                    useful error message when the two arguments are not equal.
        """
        self._type_equality_funcs[typeobj] = function

    def assertIsInstance(self, obj, cls, msg=None):
        """Same as self.assertTrue(isinstance(obj, cls)), with a nicer
        default message."""
        if not isinstance(obj, cls):
            standardMsg = '%s is not an instance of %r' % (safe_repr(obj), cls)
            self.fail(self._formatMessage(msg, standardMsg))

    def assertDictEqual(self, d1, d2, msg=None):
        self.assertIsInstance(d1, dict, 'First argument is not a dictionary')
        self.assertIsInstance(d2, dict, 'Second argument is not a dictionary')

        if d1 != d2:
            self.fail(self._formatMessage(str(), str()))

    def fail(self, msg=None):
        """Fail immediately, with the given message."""
        raise self.failureException(msg)

    def _formatMessage(self, msg, standardMsg):
        """Honour the longMessage attribute when generating failure messages.
        If longMessage is False this means:
        * Use only an explicit message if it is provided
        * Otherwise use the standard message for the assert

        If longMessage is True:
        * Use the standard message
        * If an explicit message is provided, plus ' : ' and the explicit message
        """
        if not self.longMessage:
            return msg or standardMsg
        if msg is None:
            return standardMsg
        try:
            # don't switch to '{}' formatting in Python 2.X
            # it changes the way unicode input is handled
            return '%s : %s' % (standardMsg, msg)
        except UnicodeDecodeError:
            return  '%s : %s' % (safe_repr(standardMsg), safe_repr(msg))

    def assertIsNot(self, expr1, expr2, msg=None):
        """Just like self.assertTrue(a is not b), but with a nicer default message."""
        if expr1 is expr2:
            standardMsg = 'unexpectedly identical: %s' % (safe_repr(expr1),)
            self.fail(self._formatMessage(msg, standardMsg))

    def assertIs(self, expr1, expr2, msg=None):
        """Just like self.assertTrue(a is b), but with a nicer default message."""
        if expr1 is not expr2:
            standardMsg = '%s is not %s' % (safe_repr(expr1),
                                             safe_repr(expr2))
            self.fail(self._formatMessage(msg, standardMsg))

    def assertNotEqual(self, first, second, msg=None):
        """Fail if the two objects are equal as determined by the '!='
           operator.
        """
        if not first != second:
            msg = self._formatMessage(msg, '%s == %s' % (safe_repr(first),
                                                          safe_repr(second)))
            raise self.failureException(msg)

    def _getAssertEqualityFunc(self, first, second):
        """Get a detailed comparison function for the types of the two args.

        Returns: A callable accepting (first, second, msg=None) that will
        raise a failure exception if first != second with a useful human
        readable error message for those types.
        """
        #
        # NOTE(gregory.p.smith): I considered isinstance(first, type(second))
        # and vice versa.  I opted for the conservative approach in case
        # subclasses are not intended to be compared in detail to their super
        # class instances using a type equality func.  This means testing
        # subtypes won't automagically use the detailed comparison.  Callers
        # should use their type specific assertSpamEqual method to compare
        # subclasses if the detailed comparison is desired and appropriate.
        # See the discussion in http://bugs.python.org/issue2578.
        #
        if type(first) is type(second):
            asserter = self._type_equality_funcs.get(type(first))
            if asserter is not None:
                if isinstance(asserter, str):
                    asserter = getattr(self, asserter)
                return asserter

        return self._baseAssertEqual

    def assertEqual(self, first, second, msg=None):
        """Fail if the two objects are unequal as determined by the '=='
           operator.
        """
        assertion_func = self._getAssertEqualityFunc(first, second)
        assertion_func(first, second, msg=msg)

# class TestCaseForRunInLoop(UnittestTestCaseWithState):
class TestCaseForRunInLoop(FakeTestCase):
    repeats_num: int = 1

    # @classmethod
    # def setUpClass(cls):
    #     cls.state = UnittestTestCaseState()

    # @classmethod
    # def tearDownClass(cls):
    #     cls.state.unregister()
    
    # @unittest.skip('temp skip')
    def test_dict(self):
        pifrl()
        for _ in range(self.repeats_num):
            creator: SharedMemory = SharedMemory('test_shmem', True, 200 * 1024 * 1024)

            with creator:
                val = {
                    'hello': 'world',
                }
                mapped_obj = creator.put_message(val)
                obj: Dict = creator.get_message()
                obj.update({i: str(i) for i in range(20)})
                # print(f'{val.a=}, {obj.a=}')
                self.assertIsNot(val, mapped_obj)
                self.assertIsNot(val, obj)
                self.assertIsNot(mapped_obj, obj)

                self.assertNotEqual(val, obj)
                self.assertEqual(mapped_obj, obj)
    
    def test_np_ndarray(self):
        pifrl()
        for _ in range(self.repeats_num):
            creator: SharedMemory = SharedMemory('test_shmem', True, 200 * 1024 * 1024)

            with creator:
                shape = (10, 10, 3)
                dtype = np.float64
                val = np.zeros(shape, dtype=dtype)
                mapped_obj = creator.put_message(val)
                obj = creator.get_message()

                obj[0, 0, 0] = 100.0
                obj[1, 1, 1] = 200.0
                obj[2, 2, 2] = 300.0
                obj[3, 3, 0] = 400.0
                obj[4, 4, 1] = 500.0
                obj[5, 5, 2] = 600.0
                obj[6, 6, 0] = 700.0
                obj[7, 7, 1] = 800.0
                obj[8, 8, 2] = 900.0
                obj[9, 9, 0] = 1000.0

                self.assertIsNot(val, mapped_obj)
                self.assertIsNot(val, obj)
                
                np.testing.assert_array_equal(mapped_obj, obj)
                try:
                    np.testing.assert_array_equal(val, mapped_obj)
                    assert False
                except AssertionError:
                    pass  # This is expected as we want them to be not equal
    
    def test_simple_class_inplace_obj(self):
        pifrl()
        for _ in range(self.repeats_num):
            creator: SharedMemory = SharedMemory('test_shmem', True, 200 * 1024 * 1024)

            with creator:
                val = SimpleClass(1, 2)
                mapped_obj = creator.put_message(val)
                obj = creator.get_message()
                obj.a += 1
                # print(f'{val.a=}, {obj.a=}')
                self.assertIs(val, mapped_obj)
                self.assertIsNot(val, obj)

                self.assertEqual(val, obj)
                self.assertEqual(mapped_obj, obj)
    
    def test_simple_class_obj(self):
        pifrl()
        for _ in range(self.repeats_num):
            # TODO: Fix required
            creator: SharedMemory = SharedMemory('test_shmem', True, 200 * 1024 * 1024)

            with creator:
                val = SimpleClass(1, 2)
                mapped_obj = creator.put_message(ForceGeneralObjectCopy(val))
                obj = creator.get_message()
                obj.a += 1
                # print(f'{val.a=}, {obj.a=}')
                self.assertIsNot(val, mapped_obj)
                self.assertIsNot(val, obj)

                self.assertNotEqual(val, obj)
                self.assertEqual(mapped_obj, obj)
    
    def test_dataclass_obj(self):
        pifrl()
        for _ in range(self.repeats_num):
            creator: SharedMemory = SharedMemory('test_shmem', True, 200 * 1024 * 1024)

            with creator:
                val = DataClass(1, 2)
                # print(f'{dir(val)=}')
                # pprint(val)
                # pprint(val.__dict__)
                mapped_val = creator.put_message(val)
                # print(f'{dir(mapped_val)=}')
                obj = creator.get_message()
                # print(f'{dir(obj)=}')
                obj = cast(DataClass, obj)
                obj.add_one()
                # pdi(val)
                # pdi(obj)
                # pdi(obj.__dict__)
                # print(f'{val.x=}, {obj.x=}')
                mapped_val.y = 'Yes'
                val.z = 40
                self.assertIsNot(val, obj)
                self.assertIs(val, mapped_val)
                self.assertIsNot(mapped_val, obj)

                self.assertNotEqual(val, obj)  # since internal __dict__ has no info about externally updated fields

                self.assertEqual(val.x, obj.x)
                self.assertEqual(val.y, obj.y)
                self.assertEqual(val.z, obj.z)
    
    def test_complex_dataclass_obj(self):
        pifrl()
        for _ in range(self.repeats_num):
            creator: SharedMemory = SharedMemory('test_shmem', True, 200 * 1024 * 1024)

            with creator:
                val: SomeSharedObject = SomeSharedObject(
                    some_processing_stage_control=False,
                    str_value='Hello, ',
                    data_dict={
                        'key1': 1,
                        ('key', 2): 'value2',
                        'key3': np.array([1, 2, 3], dtype=np.int32),
                    },
                    company_info=CompanyInfo(
                        company_id=1,
                        emails=('sails@company.com', 'support@company.com'),
                        websites=['http://company.com', 'http://company.org'],
                        income=1_000_000.0,
                        employees=10,
                        some_employee=Employee(
                            'John Doe', 
                            30, 
                            Positions.manager,
                            2,
                        )
                    )
                )

                # print(f'{dir(val)=}')
                # pprint(val)
                # pprint(val.__dict__)
                mapped_val = creator.put_message(val)
                # print(f'{dir(mapped_val)=}')
                obj = creator.get_message()

                obj = cast(SomeSharedObject, obj)
                obj.str_value += 'World!'
                obj.data_dict['key1'] *= 200
                obj.data_dict[('key', 2)] = obj.data_dict[('key', 2)] + ' data'
                obj.data_dict['key3'] += 10
                obj.company_info.income = 3_000_000.0
                obj.company_info.some_employee.increase_years_of_employment()
                obj.some_processing_stage_control = True

                self.assertIsNot(val, obj)
                self.assertIs(val, mapped_val)
                self.assertIsNot(mapped_val, obj)

                self.assertNotEqual(val, obj)  # since internal __dict__ has no info about externally updated fields

                self.assertEqual(val.str_value, obj.str_value)
                self.assertEqual(val.data_dict['key1'], obj.data_dict['key1'])
                self.assertEqual(val.data_dict[('key', 2)], obj.data_dict[('key', 2)])
                np.testing.assert_array_equal(val.data_dict['key3'], obj.data_dict['key3'])
                self.assertEqual(val.company_info.income, obj.company_info.income)
                self.assertEqual(val.company_info.some_employee.years_of_employment, obj.company_info.some_employee.years_of_employment)
                self.assertEqual(val.company_info.some_employee.age, obj.company_info.some_employee.age)
                self.assertEqual(val.some_processing_stage_control, obj.some_processing_stage_control)
                self.assertEqual(val.some_processing_stage_control, obj.some_processing_stage_control)
    
    def test_mutable_set_rehash(self):
        pifrl()
        for _ in range(self.repeats_num):
            creator: SharedMemory = SharedMemory('test_shmem', True, 200 * 1024 * 1024)

            with creator:
                val = set(range(1, 11))
                val2 = set(range(10, 101))
                val.update(val2)
                obj = creator.put_message(val)
                for item in val2:
                    obj.add(item)
                
                self.assertEqual(val, obj)
    
    def test_mutable_mapping_rehash(self):
        pifrl()
        for _ in range(self.repeats_num):
            creator: SharedMemory = SharedMemory('test_shmem', True, 200 * 1024 * 1024)

            with creator:
                val = {i: i for i in range(1, 11)}
                val2 = {i: i for i in range(10, 101)}
                val.update(val2)
                obj = creator.put_message(val)
                obj.update(val2)
                
                self.assertEqual(val, obj)
    
    def test_shared_memory(self):
        pifrl()
        for _ in range(self.repeats_num):
            results: List[Any] = list()
            creator: SharedMemory = SharedMemory('test_shmem', True, 200 * 1024 * 1024)

            with creator:
                results.append(creator.read_mem(creator.get_data_start_offset(), 50))
                offset, size = creator.malloc(ObjectType.tbool, 1)
                results.append(size)
                results.append(creator.read_mem(creator.get_data_start_offset(), 50))
                offset, size = creator.realloc(offset, 11)
                results.append(size)
                results.append(creator.read_mem(creator.get_data_start_offset(), 50))
                results.append(creator.free(offset))
                results.append(creator.read_mem(creator.get_data_start_offset(), 50))
                creator.set_free_memory_search_start(0)
                creator.commit_free_memory_search_start()
                results.append(creator.read_mem(creator.get_data_start_offset(), 50))
                offset, size = creator.malloc(ObjectType.tbool, 1)
                results.append(size)
                results.append(creator.read_mem(creator.get_data_start_offset(), 50))
                offset, size = creator.realloc(offset, 11)
                results.append(size)
                results.append(creator.read_mem(creator.get_data_start_offset(), 50))
                results.append(creator.free(offset))
                results.append(creator.read_mem(creator.get_data_start_offset(), 50))
                creator.set_free_memory_search_start(0)
                creator.commit_free_memory_search_start()

                values = [
                    0,
                    1,
                    999999999999,
                    -1,
                    -999999999999,
                    0.0,
                    1.0,
                    999999999999.0,
                    -1.0,
                    -999999999999.0,
                    (1,),
                    [1],
                    (1, 2),
                    [1, 2],
                    {1, 2},
                    (1, [2, 3]),
                    {1: 1, 2: 2},
                    {1: 1, 2: [2, (3,)]},
                    'test',
                    'test'.encode('utf-8'),
                    bytearray('test'.encode('utf-8')),
                    True,
                    False,
                    {
                        'test': 1,
                        'test2': 2.0,
                        'test3': 'test',
                        'test4': [
                            1, 
                            2.0, 
                            True, 
                            False, 
                            'test', 
                            'test'.encode('utf-8'), 
                            bytearray('test'.encode('utf-8')), 
                            ('hello', 1),
                            ['world', 2.0],
                            {
                                'hello': [1, (2, 3)],
                                'world': 'world',
                            }
                        ],
                    }
                ]
                
                results.append(creator.read_mem(creator.get_free_memory_search_start(), 100))
                results.append(creator.read_mem(creator.get_data_start_offset(), 500))
                
                for val in values:
                    creator.put_message(val)
                    results.append(creator.read_mem(creator.get_free_memory_search_start(), 100))
                    results.append(creator.read_mem(creator.get_data_start_offset(), 500))
                    
                    message = creator.take_message()
                    self.assertEqual(val, message)
                    results.append(message)
                    results.append(creator.read_mem(creator.get_free_memory_search_start(), 100))
                    results.append(creator.read_mem(creator.get_data_start_offset(), 500))

    def test_simple_class_obj_default(self):
        pifrl()
        for _ in range(self.repeats_num):
            # TODO: Fix required
            creator: SharedMemory = SharedMemory('test_shmem', True, 200 * 1024 * 1024)

            with creator:
                val = SimpleClass(1, 2)
                mapped_obj = creator.put_message(val)
                obj = creator.get_message()
                obj.a += 1
                # print(f'{val.a=}, {obj.a=}')
                self.assertIsNot(val, mapped_obj)
                self.assertIsNot(val, obj)

                self.assertNotEqual(val, obj)
                self.assertEqual(mapped_obj, obj)

    def test_simple_class_with_slots_obj_default(self):
        pifrl()
        for _ in range(self.repeats_num):
            # TODO: Fix required
            creator: SharedMemory = SharedMemory('test_shmem', True, 200 * 1024 * 1024)

            with creator:
                val = SimpleClassWithSlots(1, 2)
                mapped_obj = creator.put_message(val)
                obj = creator.get_message()
                obj.a += 1
                # print(f'{val.a=}, {obj.a=}')
                self.assertIsNot(val, mapped_obj)
                self.assertIsNot(val, obj)

                self.assertNotEqual(val, obj)
                self.assertEqual(mapped_obj, obj)
    
    # @unittest.skip('broken')
    def test_datetime_obj(self):
        pifrl()
        from datetime import datetime
        for _ in range(self.repeats_num):
            # TODO: Fix required
            creator: SharedMemory = SharedMemory('test_shmem', True, 200 * 1024 * 1024)

            with creator:
                val: datetime = datetime.now()
                mapped_obj: datetime = creator.put_message(val)
                obj: datetime = creator.get_message()

                self.assertIsNot(val, mapped_obj)
                self.assertIsNot(val, obj)
                self.assertIsNot(mapped_obj, obj)

                self.assertEqual(mapped_obj, val)
                self.assertEqual(mapped_obj, obj)
    
    # @unittest.skip('broken')
    def test_datetime_timedelta_obj(self):
        pifrl()
        from datetime import datetime, timedelta
        for _ in range(self.repeats_num):
            # TODO: Fix required
            creator: SharedMemory = SharedMemory('test_shmem', True, 200 * 1024 * 1024)

            with creator:
                val: timedelta = datetime.now() - datetime.now()
                mapped_obj: timedelta = creator.put_message(val)
                obj: timedelta = creator.get_message()

                self.assertIsNot(val, mapped_obj)
                self.assertIsNot(val, obj)
                self.assertIsNot(mapped_obj, obj)

                self.assertEqual(mapped_obj, val)
                self.assertEqual(mapped_obj, obj)
    
    # @unittest.skip('temp skip')
    def test_set(self):
        pifrl()
        for _ in range(self.repeats_num):
            creator: SharedMemory = SharedMemory('test_shmem', True, 200 * 1024 * 1024)

            with creator:
                val: Set = {
                    'hello',
                    'world',
                }
                mapped_obj = creator.put_message(val)
                obj: Set = creator.get_message()
                obj |= set(range(20))
                obj.discard('hello')
                # print(f'{val.a=}, {obj.a=}')
                self.assertIsNot(val, mapped_obj)
                self.assertIsNot(val, obj)
                self.assertIsNot(mapped_obj, obj)

                self.assertNotEqual(val, obj)
                print(f'{mapped_obj=}, {obj=}')
                self.assertNotEqual(mapped_obj, obj)
    
    # @unittest.skip('temp skip')
    def test_fast_set(self):
        pifrl()
        for _ in range(self.repeats_num):
            creator: SharedMemory = SharedMemory('test_shmem', True, 200 * 1024 * 1024)

            with creator:
                val: Set = FastLimitedSet({
                    'hello',
                    'world',
                })
                mapped_obj = creator.put_message(val)
                obj: Set = creator.get_message()
                obj |= set(range(20))
                obj.discard('hello')
                # print(f'{val.a=}, {obj.a=}')
                self.assertIsNot(val, mapped_obj)
                self.assertIsNot(val, obj)
                self.assertIsNot(mapped_obj, obj)

                self.assertNotEqual(val, obj)
                self.assertNotEqual(mapped_obj, obj)
                self.assertEqual(mapped_obj, val)


if __name__ == '__main__':
    test: TestCaseForRunInLoop = TestCaseForRunInLoop()
    test.test_set()



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""


__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


__all__ = ['StreamType', 'GateSecurityPolicy', 'IOCoreMemoryManagement', 'StreamManagerIOCoreMemoryManagement']


from enum import Enum
from cengal.hardware.info.cpu.versions.v_1 import cpu_info
from cengal.io.core.memory_management import IOCoreMemoryManagement
from cengal.code_flow_control.smart_values.versions import ValueExistence
from typing import Optional
from .efficient_streams_base_internal import *


class StreamType(Enum):
    general = 0
    message_based_anonymous = 1
    message_based_names = 2
    gate = 3


class GateSecurityPolicy(Enum):
    # gate will allow only or ban selected stream names to conect to this gate
    allowed = 0
    disabled = 1


class StreamManagerIOCoreMemoryManagement(IOCoreMemoryManagement):
    def __init__(self):
        super(StreamManagerIOCoreMemoryManagement, self).__init__()

        self.socket_write_fixed_buffer_size = ValueExistence(True,
                                                             cpu_info().l2_cache_size_per_virtual_core)

    def link_to(self, parent):
        super(StreamManagerIOCoreMemoryManagement, self).link_to(parent)
        try:
            self.socket_write_fixed_buffer_size = parent.socket_write_fixed_buffer_size
        except AttributeError:
            pass



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import os.path
import shutil, errno
import subprocess
import time
from ..file_system import filtered_file_list, FilteringType
import shutil
from . import upk_constants

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


def unpack_upk_files(upk_utils_dir, xcom_upk_dir, output_dir, copy_if_error=True, copy_already_unpacked_files=True):
    """
    :param upk_utils_dir:
    :param xcom_upk_dir:
    :param output_dir:
    :param copy_if_error:
    :param copy_already_unpacked_files:
    :return:
    """

    original_current_dir = os.getcwd()
    try:
        set_of_ext = {'.upk'}
        if copy_already_unpacked_files:
            set_of_ext.add('.u')
        file_list_result = filtered_file_list(xcom_upk_dir, FilteringType.including, set_of_ext)

        os.chdir(output_dir)

        for filename in file_list_result[2]:
            full_file_name = os.path.join(file_list_result[0], filename)
            call_result = subprocess.call([os.path.join(upk_utils_dir, 'DecompressLZO'), full_file_name, filename])
            print('RESULT: {}; FROM: {}; TO {}'.format(call_result, full_file_name, filename))
            if call_result == 1:  # "Package is already decompressed!"
                if copy_if_error:
                    shutil.copyfile(full_file_name, filename)
    finally:
        os.chdir(original_current_dir)


def unpack_single_upk_file(upk_utils_dir, full_file_path, output_dir, copy_if_error=True):
    original_current_dir = os.getcwd()
    try:
        os.chdir(output_dir)

        call_result = subprocess.call([os.path.join(upk_utils_dir, 'DecompressLZO'), full_file_name, filename])
        print('RESULT: {}; FROM: {}; TO {}'.format(call_result, full_file_name, filename))
        if call_result == 1:  # "Package is already decompressed!"
            if copy_if_error:
                shutil.copyfile(full_file_name, filename)
    finally:
        os.chdir(original_current_dir)


def deserialize_upk_files(upk_utils_dir, xcom_unpacked_upk_dir, output_dir):
    """
    :param upk_utils_dir:
    :param xcom_unpacked_upk_dir:
    :param output_dir:
    :return:
    """

    original_current_dir = os.getcwd()
    try:
        file_list_result = filtered_file_list(xcom_unpacked_upk_dir, FilteringType.including, {'.upk', '.u'})

        os.chdir(output_dir)

        for filename in file_list_result[2]:
            full_file_name = os.path.join(file_list_result[0], filename)
            call_result = subprocess.call([os.path.join(upk_utils_dir, 'DeserializeAll'), full_file_name])
    finally:
        os.chdir(original_current_dir)


def deserialize_single_upk_file(upk_utils_dir, upk_file_filename, output_dir):
    """
    :param upk_utils_dir:
    :param xcom_unpacked_upk_dir:
    :param output_dir:
    :return:
    """

    call_result = None
    original_current_dir = os.getcwd()
    try:
        os.chdir(output_dir)

        call_result = subprocess.call([os.path.join(upk_utils_dir, 'DeserializeAll'), upk_file_filename])
    finally:
        os.chdir(original_current_dir)

    return call_result


def patch_upk_file(upk_utils_dir, xcom_unpacked_upk_dir, patch_file, uninstall_instead_of_installation=False):
    """
    :param upk_utils_dir:
    :param xcom_unpacked_upk_dir:
    :param output_dir:
    :return:
    """

    call_result = None
    original_current_dir = os.getcwd()
    try:
        if uninstall_instead_of_installation:
            patch_file = ''.join([patch_file, upk_constants.FileExtensions.mod_uninstall_info])

        patch_file_path, patch_file_name = os.path.split(patch_file)
        os.chdir(patch_file_path)

        call_result = subprocess.call([os.path.join(upk_utils_dir, 'PatchUPK'), patch_file_name, xcom_unpacked_upk_dir])
        print(call_result)
    finally:
        os.chdir(original_current_dir)

    return call_result


def export_from_upk_to_pseudo_code(upk_utils_dir, upk_file_full_name, object_full_name, output_mod_full_file_name=None):
    """
    :param upk_utils_dir:
    :param upk_file_full_name:
    :param object_full_name:
    :param output_mod_full_file_name:
    :return: None or bytes()
    """
    mod_file_with_pseudo_code_data = None
    call_result = None
    original_current_dir = os.getcwd()
    try:
        output_mod_dir, output_mod_file_name_only = os.path.split(output_mod_full_file_name)
        os.chdir(output_mod_dir)

        upk_utility_file_name = os.path.join(upk_utils_dir, 'HexToPseudoCode')
        process = subprocess.Popen([upk_utility_file_name,
                                             upk_file_full_name,
                                             object_full_name,
                                             # '/d'
                                    ],
                                   stdout=subprocess.PIPE)
        call_result = process.returncode
        call_res_stdout, call_res_strerr = process.communicate()
        if call_result:
            # some error:
            print('ERROR:', call_res_strerr)
        else:
            # executed without errors:
            mod_file_with_pseudo_code_data = call_res_stdout
            if output_mod_full_file_name is not None:
                with open(output_mod_full_file_name, 'bw') as file:
                    file.write(call_res_stdout)
    finally:
        os.chdir(original_current_dir)

    return mod_file_with_pseudo_code_data



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from os import environ
if 'CENGAL_BUILD_IS_IN_DEBUG_MODE' in environ:
    import debugpy
    debugpy.listen(("0.0.0.0", 5678))
    debugpy.wait_for_client()
    debugpy.breakpoint()


environ['CENGAL_IS_IN_BUILD_MODE'] = 'True'


import setuptools
# from setuptools import Extension, setup
# import platform


setuptools._install_setup_requires({'setup_requires': ['Cython>=0.29.34', 'py-cpuinfo', 'charset-normalizer; python_version < "3.13"', 'PyBindGen']})


from cengal_setup_scripts.install_required_packages.install_packages import install_bundled, get_pypi_requirements_list, get_remote_requirements_list
# install_bundled()
from cengal_setup_scripts.find_and_prepare_cython_modules import find_and_prepare_cython_modules, build, build_ext, sdist, \
    find_good_packages, find_package_data
from cengal.file_system.path_manager import path_relative_to_src


with open(path_relative_to_src('README.md'), 'r') as fh:
    long_description = fh.read()


setup_requires = [
    "charset-normalizer; python_version < '3.13'", 
    "chardet; python_version < '3.11'",
    "py-cpuinfo",
    "Cython>=0.29.34",
    "patch-ng; python_version >= '3.12' or (platform_python_implementation == 'PyPy' and platform_system == 'Windows')",
    "typing_extensions",
    "PyBindGen",
]
pypi_requirements_list = get_pypi_requirements_list()
if 'patch-ng' in pypi_requirements_list:
    pypi_requirements_list.remove('patch-ng')
    pypi_requirements_list.insert(0, 'patch-ng')

remote_requirements_list = get_remote_requirements_list()


from cengal.system import CENGAL_IS_IN_BUILD_MODE
from cengal.build_tools.current_compiler import compiler_type
from cengal.build_tools.prepare_cflags import prepare_given_cflags, prepare_cflags, concat_cflags, prepare_compile_time_env


prepare_given_cflags({
    'CENGAL_IS_IN_BUILD_MODE': (False, CENGAL_IS_IN_BUILD_MODE),
    'CENGAL_BUILD_IS_IN_DEBUG_MODE': (False, 'CENGAL_BUILD_IS_IN_DEBUG_MODE' in environ),
})


if 'CENGAL_BUILD_IS_IN_DEBUG_MODE' in environ:
    debugpy.breakpoint()


packages_data_dict, manifest_included_files = find_package_data()
setuptools.setup(
    name='cengal_light',
    version=__version__,
    author='ButenkoMS',
    author_email=__email__,
    description='General purpose library',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/FI-Mihej/Cengal',
    license=__license__,
    keywords='async loop, coroutine, async wxPython, async Qt, async Tkinter, bytecode manipulation, introspection, text parsing',
    py_modules=['cengal'],
    # package_dir={'': path_relative_to_src('')},
    packages=find_good_packages(),
    package_data=packages_data_dict,
    setup_requires=setup_requires,
    # setup_requires=pypi_requirements_list,
    entry_points={
        'console_scripts': [
            'cengal__package__create = cengal.build_tools.packages.create:main',
            'cengal__package__add_project = cengal.build_tools.packages.add_project:main',
        ],
    },
    extras_require={
        'setuprequires': setup_requires,
        'entities__asm_functions': [
            'py-cpuinfo>=7.0.0',
        ],
        'hardware__info__cpu__versions__v_0': [
            'typing_extensions',
            'py-cpuinfo',
        ],
        'hardware__info__cpu__versions__v_1': [
            'typing_extensions',
            'py-cpuinfo',
            "psutil; platform_python_implementation != 'PyPy' or platform_system != 'Windows'",
        ],
        'cengal__time_management__relative_time__relativedelta': [
            'typing_extensions',
            'python-dateutil',
        ],
        'cengal__web_tools__help_tools': [
            'typing_extensions',
            'requests',
            "http_parser; python_version < '3.11'",
        ],
        'cengal__user_interface__console__chooser': [
            'typing_extensions',
            'colorama',
        ],
        'cengal__user_interface__console__colorama_helpers': [
            'typing_extensions',
            'colorama',
        ],
        'cengal__parallel_execution__coroutines': [
            'typing_extensions',
            'greenlet',
        ],
        'cengal__parallel_execution__coroutines__coro_standard_services__db': [
            'typing_extensions',
            "patch-ng; ('ARM' in platform_machine) or ((python_version >= '3.12' and platform_python_implementation == 'CPython') or (platform_python_implementation == 'PyPy' and ((platform_system == 'Windows' and python_version != '3.9' and python_version != '2.7') or (platform_system == 'Darwin' and python_version != '2.7'))))",
            "lmdb; python_version != '3.10' or platform_python_implementation != 'PyPy' or platform_system != 'Windows'",
        ],
        'cengal__parallel_execution__coroutines__coro_standard_services__lmdb': [
            'typing_extensions',
            "patch-ng; ('ARM' in platform_machine) or ((python_version >= '3.12' and platform_python_implementation == 'CPython') or (platform_python_implementation == 'PyPy' and ((platform_system == 'Windows' and python_version != '3.9' and python_version != '2.7') or (platform_system == 'Darwin' and python_version != '2.7'))))",
            "lmdb; python_version != '3.10' or platform_python_implementation != 'PyPy' or platform_system != 'Windows'",
        ],
        'cengal__parallel_execution__coroutines__coro_standard_services__log': [
            'typing_extensions',
            "patch-ng; ('ARM' in platform_machine) or ((python_version >= '3.12' and platform_python_implementation == 'CPython') or (platform_python_implementation == 'PyPy' and ((platform_system == 'Windows' and python_version != '3.9' and python_version != '2.7') or (platform_system == 'Darwin' and python_version != '2.7'))))",
            "lmdb; python_version != '3.10' or platform_python_implementation != 'PyPy' or platform_system != 'Windows'",
        ],
        'cengal__parallel_execution__coroutines__coro_standard_services__loop_yield': [
            'typing_extensions',
            'async_generator',
        ],
        'cengal__parallel_execution__coroutines__coro_standard_services__read_write_locker': [
            'typing_extensions',
            'async_generator',
        ],
        'cengal__parallel_execution__coroutines__coro_standard_services__tkinter': [
            'typing_extensions',
            'async_generator',
        ],
        'cengal__parallel_execution__coroutines__coro_tools__loop_administration__admin_tk': [
            'typing_extensions',
            'ttkbootstrap',
            'pprintpp',
        ],
        'cengal__data_manipulation__serialization': [
            'typing_extensions',
            "orjson; platform_python_implementation != 'PyPy'",
            'simplejson', 'cbor', 'cbor2', 'ujson', 'cloudpickle', 
            "msgpack-python; python_version < '3.8' and platform_python_implementation == 'CPython'",
            "msgpack; python_version >= '3.8' and platform_python_implementation == 'CPython'",
            "msgpack-pypy; platform_python_implementation == 'PyPy'",
            "msgspec; python_version < '3.13' and platform_python_implementation == 'CPython'",
        ],
        'cengal__user_interface__gui__tkinter': [
            'typing_extensions',
            'ttkbootstrap',
        ],
        'cengal__text_processing__encoding_detection': [
            'typing_extensions',
            "charset-normalizer; python_version < '3.13'", 
            "chardet; python_version < '3.11'",
        ],
        'cengal__file_system__app_fs_structure__app_dir_path': [
            'typing_extensions',
            "pyobjc; platform_system == 'Darwin' and platform_python_implementation == 'CPython'",
            "pywin32; platform_system == 'Windows' and platform_python_implementation == 'CPython'",
        ],
        'cengal__build_tools__packages__create': [
            'typing_extensions',
            "aiodns",
            "aiohttp",
        ],
        'cengal__build_tools__gather_docs': [
            'pdoc',
        ],
        'cengal__build_tools__build_extensions': [
            'PyBindGen',
        ],
        'full': [
            # 'PyBindGen',
            'typing_extensions',
            "msgspec; python_version < '3.13' and platform_python_implementation == 'CPython'",
            "pyobjc; platform_system == 'Darwin' and platform_python_implementation == 'CPython'",
            "pywin32; platform_system == 'Windows' and platform_python_implementation == 'CPython'",
            'pdoc',
            "psutil; platform_python_implementation != 'PyPy' or platform_system != 'Windows'",
            "orjson; platform_python_implementation != 'PyPy'",
            "patch-ng; ('ARM' in platform_machine) or ((python_version >= '3.12' and platform_python_implementation == 'CPython') or (platform_python_implementation == 'PyPy' and ((platform_system == 'Windows' and python_version != '3.9' and python_version != '2.7') or (platform_system == 'Darwin' and python_version != '2.7'))))",
            "lmdb; python_version != '3.10' or platform_python_implementation != 'PyPy' or platform_system != 'Windows'",
            'py-cpuinfo>=7.0.0', 'python-dateutil', 'greenlet', 'async_generator',
            'ttkbootstrap', 'pprintpp', 'requests', 'simplejson', 'cbor', 'cbor2', 'ujson', 'cloudpickle', 
            "http_parser; python_version < '3.11'",
            "msgpack-python; python_version < '3.8' and platform_python_implementation == 'CPython'",
            "msgpack; python_version >= '3.8' and platform_python_implementation == 'CPython'",
            "msgpack-pypy; platform_python_implementation == 'PyPy'",
            "charset-normalizer; python_version < '3.13'", 
            "chardet; python_version < '3.11'",
            "aiodns",
            "aiohttp",
        ], 
        'all': [
            # 'PyBindGen',
            'typing_extensions',
            "msgspec; python_version < '3.13' and platform_python_implementation == 'CPython'",
            "pyobjc; platform_system == 'Darwin' and platform_python_implementation == 'CPython'",
            "pywin32; platform_system == 'Windows' and platform_python_implementation == 'CPython'",
            'pdoc',
            "psutil; platform_python_implementation != 'PyPy' or platform_system != 'Windows'",
            "orjson; platform_python_implementation != 'PyPy'",
            "patch-ng; ('ARM' in platform_machine) or ((python_version >= '3.12' and platform_python_implementation == 'CPython') or (platform_python_implementation == 'PyPy' and ((platform_system == 'Windows' and python_version != '3.9' and python_version != '2.7') or (platform_system == 'Darwin' and python_version != '2.7'))))",
            "lmdb; python_version != '3.10' or platform_python_implementation != 'PyPy' or platform_system != 'Windows'",
            'py-cpuinfo>=7.0.0', 'python-dateutil', 'greenlet', 'async_generator',
            'ttkbootstrap', 'pprintpp', 'requests', 'simplejson', 'cbor', 'cbor2', 'ujson', 'cloudpickle', 
            "http_parser; python_version < '3.11'",
            "msgpack-python; python_version < '3.8' and platform_python_implementation == 'CPython'",
            "msgpack; python_version >= '3.8' and platform_python_implementation == 'CPython'",
            "msgpack-pypy; platform_python_implementation == 'PyPy'",
            "charset-normalizer; python_version < '3.13'", 
            "chardet; python_version < '3.11'",
            "aiodns",
            "aiohttp",
        ], 
    },
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Environment :: MacOS X',
        'Environment :: Web Environment',
        'Environment :: WebAssembly',
        'Environment :: Win32 (MS Windows)',
        'Environment :: X11 Applications',
        'Framework :: AsyncIO',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Manufacturing',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Telecommunications Industry',
        'Operating System :: Android',
        'Operating System :: MacOS',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: Microsoft :: Windows :: Windows 7',
        'Operating System :: Microsoft :: Windows :: Windows 8',
        'Operating System :: Microsoft :: Windows :: Windows 8.1',
        'Operating System :: Microsoft :: Windows :: Windows 10',
        'Operating System :: Microsoft :: Windows :: Windows 11',
        'Operating System :: Microsoft :: Windows :: Windows Vista',
        'Operating System :: Microsoft :: Windows :: Windows XP',
        'Operating System :: Microsoft :: Windows :: Windows Server 2008',
        'Operating System :: Microsoft :: Windows :: Windows Server 2003',
        'Operating System :: Microsoft :: Windows :: Windows NT/2000',
        'Operating System :: POSIX',
        'Operating System :: POSIX :: BSD',
        'Operating System :: POSIX :: BSD :: OpenBSD',
        'Operating System :: POSIX :: BSD :: FreeBSD',
        'Operating System :: POSIX :: BSD :: NetBSD',
        'Operating System :: POSIX :: Linux',
        'Operating System :: Unix',
        'Operating System :: iOS',
        'Programming Language :: C',
        'Programming Language :: Cython',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: Implementation',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Programming Language :: Python :: Implementation :: MicroPython',
        'Topic :: Communications',
        'Topic :: Database',
        'Topic :: Database :: Front-Ends',
        'Topic :: Database :: Database Engines/Servers',
        'Topic :: Desktop Environment',
        'Topic :: Games/Entertainment',
        'Topic :: Internet',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Browsers',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content :: CGI Tools/Libraries',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content :: Page Counters',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content :: Wiki',
        'Topic :: Internet :: WWW/HTTP :: HTTP Servers',
        'Topic :: Internet :: WWW/HTTP :: Indexing/Search',
        'Topic :: Multimedia',
        'Topic :: Multimedia :: Graphics',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Scientific/Engineering :: Visualization',
        'Topic :: Software Development',
        'Topic :: Software Development :: Assemblers',
        'Topic :: Software Development :: Build Tools',
        'Topic :: Software Development :: Code Generators',
        'Topic :: Software Development :: Compilers',
        'Topic :: Software Development :: Debuggers',
        'Topic :: Software Development :: Disassemblers',
        'Topic :: Software Development :: Internationalization',
        'Topic :: Software Development :: Interpreters',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: Localization',
        'Topic :: Software Development :: Object Brokering',
        'Topic :: Software Development :: Pre-processors',
        'Topic :: Software Development :: Testing',
        'Topic :: Software Development :: Testing :: Mocking',
        'Topic :: Software Development :: Testing :: Unit',
        'Topic :: Software Development :: User Interfaces',
        'Topic :: Software Development :: Version Control :: Git',
        'Topic :: Software Development :: Widget Sets',
        'Topic :: System',
        'Topic :: System :: Benchmark',
        'Topic :: System :: Clustering',
        'Topic :: System :: Distributed Computing',
        'Topic :: System :: Filesystems',
        'Topic :: System :: Hardware',
        'Topic :: System :: Hardware :: Symmetric Multi-processing',
        'Topic :: System :: Installation/Setup',
        'Topic :: System :: Logging',
        'Topic :: System :: Monitoring',
        'Topic :: System :: Networking',
        'Topic :: System :: Networking :: Monitoring',
        'Topic :: System :: Networking :: Monitoring :: Hardware Watchdog',
        'Topic :: System :: Shells',
        'Topic :: Text Processing',
        'Topic :: Text Processing :: Filters',
        'Topic :: Text Processing :: General',
        'Topic :: Text Processing :: Indexing',
        'Topic :: Text Processing :: Linguistic',
        'Topic :: Text Processing :: Markup',
        'Topic :: Utilities',
        'Typing :: Typed',
        # 'Private :: Do Not Upload',
    ],
    python_requires='>=3.8',
    ext_modules=find_and_prepare_cython_modules(),
    # ext_modules=cythonize(
    #     find_and_prepare_cython_modules(),
    #     compiler_directives={'language_level': '3'},
    #     compile_time_env = prepare_compile_time_env(),
    # ),
    cmdclass={
        # 'build_ext': build_ext,
        'build': build,
        'sdist': sdist,
    },
)



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = ['UAProf', 'is_wap_browser', 'BrowserViewState', 'browser_view_state', 'ClientDevicePlatform', 
           'client_device_platform', 'client_device_platform_version', 'client_device_model', 'client_device_arch', 
           'MobileDeviceType', 'mobile_device_type_by_user_agent', 'ClientDeviceType', 'client_device_type', 
           'ClientViewType', 'client_view_type']


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from enum import Enum, IntEnum
from typing import Mapping, Optional
from cengal.web_tools.detect_browsers_host_device_type.by_http_user_agent import is_mobile_or_tablet_browser


UAProf = {
    'x-wap-profile'.casefold(),
    '19-Profile'.casefold(),
    'WAP-Profile'.casefold(),
}


def is_wap_browser(headers: Mapping) -> bool:
    return bool(UAProf.intersection(headers.keys()))


class BrowserViewState(str, Enum):
    desktop = '?0'
    mobile = '?1'
    unknown = 'unknown'


def browser_view_state(headers: Mapping) -> BrowserViewState:
    return BrowserViewState(headers.get('Sec-CH-UA-Mobile'.casefold(), 'unknown'))


class ClientDevicePlatform(str, Enum):
    unknown = 'unknown'
    windows = 'Windows'
    macos = 'macOS'
    linux = 'Linux'
    ios = 'iOS'
    chromiumos = 'Chromium OS'
    chromeos = 'Chrome OS'
    android = 'Android'


def client_device_platform(headers: Mapping) -> ClientDevicePlatform:
    return ClientDevicePlatform(headers.get('Sec-CH-UA-Platform'.casefold(), 'unknown'))


def client_device_platform_version(headers: Mapping) -> Optional[str]:
    return headers.get('Sec-CH-UA-Platform-Version'.casefold(), None)


def client_device_model(headers: Mapping) -> Optional[str]:
    return headers.get('Sec-CH-UA-Model'.casefold(), None)


def client_device_arch(headers: Mapping) -> Optional[str]:
    return headers.get('Sec-CH-UA-Arch'.casefold(), None)


class MobileDeviceType(str, Enum):
    unknown = 'unknown'
    mobile = 'mobile'
    tablet = 'tablet'


def mobile_device_type_by_user_agent(headers: Mapping) -> MobileDeviceType:
    mobile, tablet = is_mobile_or_tablet_browser(headers.get('User-Agent'.casefold(), ''))
    if tablet:
        return MobileDeviceType.tablet
    elif mobile:
        return MobileDeviceType.mobile
    else:
        return MobileDeviceType.unknown


class ClientDeviceType(str, Enum):
    desktop = 'desktop'
    laptop = 'laptop'
    tablet = 'tablet'
    mobile = 'mobile'


def client_device_type(headers: Mapping) -> ClientDeviceType:
    if is_wap_browser(headers):
        return ClientDeviceType.mobile
    else:
        cdp: ClientDevicePlatform = client_device_platform(headers)
        if ClientDevicePlatform.unknown == cdp:
            mdt: MobileDeviceType = mobile_device_type_by_user_agent(headers)
            if MobileDeviceType.tablet == mdt:
                return ClientDeviceType.tablet
            elif MobileDeviceType.mobile == mdt:
                return ClientDeviceType.mobile
            else:
                return ClientDeviceType.desktop
        elif cdp in {ClientDevicePlatform.ios, ClientDevicePlatform.android}:
            mdt: MobileDeviceType = mobile_device_type_by_user_agent(headers)
            if MobileDeviceType.tablet == mdt:
                return ClientDeviceType.tablet
            else:
                return ClientDeviceType.mobile
        else:
            return ClientDeviceType.desktop


class ClientViewType(str, Enum):
    desktop = 'desktop'
    laptop = 'laptop'
    tablet = 'tablet'
    mobile = 'mobile'


def client_view_type(headers: Mapping) -> ClientViewType:
    bvs: BrowserViewState = browser_view_state(headers)
    if BrowserViewState.desktop == bvs:
        return ClientViewType.desktop
    elif BrowserViewState.mobile == bvs:
        mdt: MobileDeviceType = mobile_device_type_by_user_agent(headers)
        if MobileDeviceType.tablet == mdt:
            return ClientViewType.tablet
        else:
            return ClientViewType.mobile
    elif BrowserViewState.unknown == bvs:
        if is_wap_browser(headers):
            return ClientDeviceType.mobile
        else:
            mdt: MobileDeviceType = mobile_device_type_by_user_agent(headers)
            if MobileDeviceType.tablet == mdt:
                return ClientViewType.tablet
            elif MobileDeviceType.mobile == mdt:
                return ClientViewType.mobile
            else:
                return ClientViewType.desktop
    else:
        raise ValueError(f'Unknown BrowserViewState: {bvs}')



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = [
    'escape_text', 
    'handle_spaces', 
    'escape_param', 
    'prepare_params', 
    'prepare_py_params', 
    'prepare_command', 
    'prepare_py_command', 
]


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from cengal.system import OS_TYPE

import os
import sys
from typing import Sequence, Union, List


def escape_text(text: str) -> str:
    """Resulting text can be used safely inside a brackets:

    param = f'--define:MyVar="{escape_text(param)}"'

    Args:
        text (str): _description_

    Returns:
        str: _description_
    """
    if 'Windows' == OS_TYPE:
        text = text.replace('"', '""')
    else:
        text = text.translate(str.maketrans({'\\': '\\\\', '"': '\\"', '$': '\\$', '`': '\\`'}))

    return text


def handle_spaces(text: str) -> str:
    if ' ' in text:
        if 'Windows' == OS_TYPE:
            # text = '"{}"'.format(text)
            need_to_wrap: bool = False
            start = 0
            while True:
                space_pos = text.find(' ', start)
                if -1 >= space_pos:
                    break

                lbracket_pos = text.find('"', start)
                rbracket_pos = text.find('"', space_pos + 1)
                if (-1 < lbracket_pos) and (-1 < rbracket_pos) and (lbracket_pos < space_pos < rbracket_pos):
                    start = rbracket_pos + 1
                    continue
                else:
                    need_to_wrap = True
                    break
            
            if need_to_wrap:
                text = text.replace('"', '""')
                text = '"{}"'.format(text)
        else:
            # text = text.replace(' ', '\\ ')
            unescaped_spaces: List[int] = list()
            probably_unescaped_spaces: List[int] = list()
            escaped: bool = False
            lbracket_found: bool = False
            for index, char in enumerate(text):
                if escaped:
                    escaped = False
                    continue

                if '\\' == char:
                    escaped = True
                elif ' ' == char:
                    if lbracket_found:
                        probably_unescaped_spaces.append(index)
                    else:
                        unescaped_spaces.append(index)
                elif '"' == char:
                    if lbracket_found:
                        probably_unescaped_spaces.clear()
                        lbracket_found = False
                    else:
                        lbracket_found = True

            unescaped_spaces.extend(probably_unescaped_spaces)
            if unescaped_spaces:
                characters: List[str] = list(text)
                for space_pos in reversed(unescaped_spaces):
                    characters.insert(space_pos, '\\')
                
                text = ''.join(characters)

    return text


def escape_param(text: str) -> str:
    return handle_spaces(escape_text(text))


def prepare_params(params: Sequence[str]) -> List[str]:
    result: List[str] = list()
    for param in params:
        result.append(handle_spaces(param))
    
    return result


def prepare_py_params(params: Sequence[str]) -> List[str]:
    result: List[str] = [handle_spaces(sys.executable)]
    for param in params:
        result.append(handle_spaces(param))
    
    return result


def prepare_command(command: str, params: Union[str, Sequence[str]] = None) -> str:
    command = handle_spaces(command)
    params_has_spaces_under_windows: bool = False
    if (params is not None) and (not isinstance(params, str)):
        new_params: List[str] = list()
        for param in params:
            if ('Windows' == OS_TYPE) and ' ' in param:
                params_has_spaces_under_windows = True
                
            new_params.append(handle_spaces(param))

        params = ' '.join(new_params)
    
    if params:
        command = '{} {}'.format(command, params)
        if params_has_spaces_under_windows:
            command = '"{}"'.format(command)
    
    return command


def prepare_py_command(params: Union[str, Sequence[str]] = None) -> str:
    return prepare_command(sys.executable, params)



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = [
    'ASharedMemoryContextManager',
    'ASMCM',
    'ASharedMemoryManager',
    'ASMM',
]


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from cengal.hardware.memory.shared_memory import *
from cengal.parallel_execution.asyncio.atasks import create_task_awaitable
from cengal.time_management.sleep_tools import get_usable_min_sleep_interval
from cengal.code_flow_control.smart_values import ValueHolder

import asyncio
from typing import Any, Coroutine


class ASharedMemoryContextManager:
    def __init__(self, ashared_memory_management: 'ASharedMemoryManager', if_has_messages: Optional[bool] = None):
        self.ashared_memory_management: 'ASharedMemoryManager' = ashared_memory_management
        self._if_has_messages_original_value: Optional[bool] = if_has_messages
        self._if_has_messages_value: Optional[bool] = None
        self.if_has_messages_value = if_has_messages
        self.shared_memory: SharedMemory = self.ashared_memory_management.shared_memory
        self.shared_memory_holder: ValueHolder[SharedMemory] = ValueHolder(True, self.shared_memory)
        self.client_iteration_done: asyncio.Future = None
        self._entered: bool = False

    async def __aenter__(self) -> ValueHolder[SharedMemory]:
        self._entered = True
        if self.if_has_messages_value:
            await self.ashared_memory_management.has_messages()
        
        line_ready: asyncio.Future = self.ashared_memory_management.get_in_line()
        self.client_iteration_done = await line_ready
        self.shared_memory_holder.existence = True
        return self.shared_memory_holder
    
    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self.client_iteration_done.set_result(bool(self.shared_memory_holder))
        self.client_iteration_done = None
        self.if_has_messages_value = self._if_has_messages_original_value
        self._entered = False

    @property
    def save_cpu_time(self):
        return self.ashared_memory_management.save_cpu_time
    
    @save_cpu_time.setter
    def save_cpu_time(self, value: bool):
        self.ashared_memory_management.save_cpu_time = value
    
    @property
    def if_has_messages_value(self) -> Optional[bool]:
        return self._if_has_messages_value
    
    @if_has_messages_value.setter
    def if_has_messages_value(self, value: Optional[bool]):
        self._if_has_messages_value = False if value is None else value
    
    @property
    def _loop(self):
        return self.ashared_memory_management._loop
    
    def close(self) -> Optional[asyncio.Future]:
        return self.ashared_memory_management.close()
    
    async def aclose(self):
        await self.ashared_memory_management.aclose()
    
    async def has_messages(self):
        if self._entered:
            raise RuntimeError('You cannot wait for `has_messages()` while inside the context manager. This could cause the program to hang, as it might prevent other processes from taking their turn and adding new messages.')
        
        return await self.ashared_memory_management.has_messages()
    
    def if_has_messages(self, value: bool = True):
        self.if_has_messages_value = value
        return self
    
    def anyway(self):
        self.if_has_messages_value = False
        return self


ASMCM = ASharedMemoryContextManager


class ASharedMemoryManager:
    def __init__(self, shared_memory: SharedMemory, save_cpu_time: bool = False, time_limit: Optional[RationalNumber] = None):
        self.shared_memory: SharedMemory = shared_memory
        self._save_cpu_time: bool = None
        self._asleep_func: Coroutine = None
        self.save_cpu_time = save_cpu_time
        self.time_limit: Optional[RationalNumber] = time_limit
        self._loop: asyncio.AbstractEventLoop = None
        self.requests_from_clients: List[asyncio.Future] = list()
        self.clients_readyness: List[asyncio.Future] = list()
        self.closed: bool = False
        self.close_requested: bool = False
        self.close_requests: List[asyncio.Future] = list()
        self.has_messages_requests: List[asyncio.Future] = list()
        self.worker_task: asyncio.Task = None
    
    async def __aenter__(self) -> 'ASharedMemoryManager':
        self.closed = False
        self.run()
        return self
    
    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.aclose()

    async def _sleep_till_next_loop_iteration(self):
        await asyncio.sleep(0)
        mm_pause()
    
    async def _sleep_for_minimal_time(self):
        await asyncio.sleep(get_usable_min_sleep_interval())
        mm_pause()
    
    @property
    def save_cpu_time(self) -> bool:
        return self._save_cpu_time
    
    @save_cpu_time.setter
    def save_cpu_time(self, value: bool):
        self._save_cpu_time = value
        if value:
            self.shared_memory._asleep_func = self._asleep_func = self._sleep_for_minimal_time
        else:
            self.shared_memory._asleep_func = self._asleep_func = self._sleep_till_next_loop_iteration

    async def aclose(self) -> None:
        if self.closed:
            return
        
        self.close_requested = True
        close_request = self._loop.create_future()
        self.close_requests.append(close_request)
        await close_request

    def close(self) -> Optional[asyncio.Future]:
        if self.closed:
            return None

        self.close_requested = True
        close_request = self._loop.create_future()
        self.close_requests.append(close_request)
        return close_request
    
    def run(self):
        if self.worker_task:
            return
        
        self._loop = asyncio.get_event_loop()
        self.worker_task = create_task_awaitable(self.worker())
    
    def get_in_line(self) -> asyncio.Future:
        get_in_line_future = self._loop.create_future()
        self.requests_from_clients.append(get_in_line_future)
        return get_in_line_future

    async def has_messages(self):
        has_messages_future = self._loop.create_future()
        self.has_messages_requests.append(has_messages_future)
        return await has_messages_future
    
    def __call__(self, if_has_messages: Optional[bool] = None) -> ASharedMemoryContextManager:
        return ASharedMemoryContextManager(self, if_has_messages)
    
    async def worker(self):
        try:
            async with self.shared_memory:
                if self.shared_memory.create:
                    await self.shared_memory.await_consumer_ready(self.time_limit)
                else:
                    await self.shared_memory.ainit_consumer(self.time_limit)
                
                while not self.close_requested:
                    if self.requests_from_clients or self.has_messages_requests:
                        if self.shared_memory.get_in_line():
                            try:
                                if self.shared_memory.has_messages():
                                    has_messages_requests_buff = self.has_messages_requests
                                    self.has_messages_requests = list()
                                    for has_messages_request in has_messages_requests_buff:
                                        has_messages_request.set_result(True)
                                
                                requests_from_clients_buff = self.requests_from_clients
                                self.requests_from_clients = list()
                                for request_from_client in requests_from_clients_buff:
                                    client_ready_future = self._loop.create_future()
                                    request_from_client.set_result(client_ready_future)
                                    self.clients_readyness.append(client_ready_future)
                                
                                client_results = await asyncio.gather(*self.clients_readyness, return_exceptions=True)
                                self.clients_readyness.clear()
                            finally:
                                self.shared_memory.release()
                            
                            if any(client_results):
                                await asyncio.sleep(0)
                            else:
                                await asyncio.sleep(get_usable_min_sleep_interval())
                    else:
                        await self._asleep_func()
        finally:
            self.closed = True
            for close_request in self.close_requests:
                close_request.set_result(None)
            
            self.close_requests.clear()
            self.close_requested = False
            self.worker_task = None


ASMM = ASharedMemoryManager



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = ['AnAppropriateValues', 'AllPossibleValues', 'AnAppropriateContainers', 'data_2_tuple', 'TreeTraversalType', 'TreeTraversal']


from typing import Union, Dict, Set, List, Tuple, Optional, Callable
from enum import Enum
from collections import deque
from cengal.code_flow_control.smart_values import ValueExistence
from cengal.data_containers.stack import StackItem, TreeStackItem, Stack, recursion_insureness

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


AnAppropriateValues = Union[str, bytes, int, float, None]
AllPossibleValues = Union['AnAppropriateContainers', AnAppropriateValues]
AnAppropriateContainers = Union[Dict[AnAppropriateValues, AllPossibleValues],
                                Set[AllPossibleValues],
                                List[AllPossibleValues],
                                Tuple[AllPossibleValues]]


def data_2_tuple(data: AllPossibleValues) -> AllPossibleValues:
    if isinstance(data, (dict, tuple, list, set, deque)):
        if isinstance(data, dict):
            return tuple(data.values())
        elif isinstance(data, (tuple, list, deque)):
            return data
        elif isinstance(data, set):
            return tuple(data)
    else:
        return data


class TreeTraversalType(Enum):
    recursive = 0
    stack_based = 1
    recursive_with_stack_based_insureness = 2


class TreeTraversal:
    def __init__(self,
                 tree: AnAppropriateContainers=None,
                 on_node: Optional[Callable]=None,
                 on_child: Optional[Callable]=None,
                 on_switched_to_stack_based_implementation: Optional[Callable]=None
                 ):
        self._tree = tree  # type: AnAppropriateContainers

        self._callback_on_node = on_node  # type: Optional[Callable]
        self._callback_on_child = on_child  # type: Optional[Callable]
        self._on_switched_to_stack_based_implementation = \
            on_switched_to_stack_based_implementation  # type: Optional[Callable]

        self._last_tree_node = ValueExistence(False, None)
        self._last_tree_child = ValueExistence(False, None)
        self._empty_child = ValueExistence(False, None)

    def set_callback_on_node(self, functor: Optional[Callable]):
        self._callback_on_node = functor

    def set_callback_on_child(self, functor: Optional[Callable]):
        self._callback_on_child = functor

    def set_callback_on_both_node_and_child(self, functor: Optional[Callable]):
        self._callback_on_node = self._callback_on_child = functor

    def set_callback_on_switched_to_stack_based_implementation(self, functor: Optional[Callable]):
        self._on_switched_to_stack_based_implementation = functor

    @property
    def tree(self) -> AnAppropriateContainers:
        return self._tree

    @tree.setter
    def tree(self, tree: AnAppropriateContainers):
        self._tree = tree

    def _traversal_recursive(self):
        tree_tuple = data_2_tuple(self.tree)
        if isinstance(self.tree, dict):
            self._last_tree_node.value = self.tree
            if self._callback_on_node:
                self._callback_on_node(self._last_tree_node, self._empty_child, None)
            self._last_tree_node.value = tree_tuple
        else:
            self._last_tree_child.value = self.tree
            if self._callback_on_child:
                self._callback_on_child(self._last_tree_node, self._last_tree_child, None)

        self._traversal_recursive_impl(tree_tuple)

    def _traversal_recursive_impl(self, tree: AnAppropriateContainers):
        if isinstance(tree, (tuple, list, deque)):
            index = 0
            for item in tree:
                next_child = item
                next_child_tuple = data_2_tuple(next_child)
                if isinstance(next_child, (dict, set, tuple, list, deque)):
                    self._last_tree_node.value = next_child
                    if self._callback_on_node:
                        self._callback_on_node(self._last_tree_node, self._empty_child, None)
                    self._last_tree_node.value = next_child_tuple
                else:
                    self._last_tree_child.value = next_child
                    if self._callback_on_child:
                        self._callback_on_child(self._last_tree_node, self._last_tree_child, index)

                self._traversal_recursive_impl(next_child_tuple)
                index += 1

    def _traversal_stack_based(self):
        tree_tuple = data_2_tuple(self.tree)
        if isinstance(self.tree, dict):
            self._last_tree_node.value = self.tree
            if self._callback_on_node:
                self._callback_on_node(self._last_tree_node, self._empty_child, None)
            self._last_tree_node.value = tree_tuple
        else:
            self._last_tree_child.value = self.tree
            if self._callback_on_child:
                self._callback_on_child(self._last_tree_node, self._last_tree_child, None)

        stack = Stack(TreeStackItem(tree_tuple))
        while stack:
            if stack.top().child is None:
                if isinstance(stack.top().node, (tuple, list, deque)):
                    if stack.top().node:
                        stack.top().child = -1
                        continue
                    else:
                        stack.pop()
                        continue
                else:
                    stack.pop()
                    continue
            else:
                stack.top().child += 1
                if len(stack.top().node) > stack.top().child:
                    next_child = stack.top().node[stack.top().child]
                    next_child_tuple = data_2_tuple(next_child)
                    if isinstance(next_child, (dict, set, tuple, list, deque)):
                        self._last_tree_node.value = next_child
                        if self._callback_on_node:
                            self._callback_on_node(self._last_tree_node, self._empty_child, None)
                        self._last_tree_node.value = next_child_tuple
                    else:
                        self._last_tree_child.value = next_child
                        if self._callback_on_child:
                            self._callback_on_child(self._last_tree_node, self._last_tree_child, stack.top().child)

                    stack.push(TreeStackItem(next_child_tuple))
                    continue
                else:
                    stack.pop()
                    continue

    def __call__(self, traversal_type: TreeTraversalType=TreeTraversalType.recursive_with_stack_based_insureness):
        if TreeTraversalType.recursive == traversal_type:
            self._traversal_recursive()
        elif TreeTraversalType.stack_based == traversal_type:
            self._traversal_stack_based()
        else:
            recursion_insureness(self._traversal_recursive,
                                 self._traversal_stack_based,
                                 self._on_switched_to_stack_based_implementation)



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = ['module_dir', 'module_rel_path', 'module_import_str', 'current_module_dir', 'current_module_rel_path', 'current_module_import_str']


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from cengal.introspection.inspect import entity_owning_module_info_and_owning_path, get_module_importable_str_and_path, frame
from cengal.introspection.inspect import entity_name
from cengal.file_system.path_manager import get_relative_path_part
from cengal.text_processing.text_processing import removesuffix
import os
from typing import Optional


def module_dir(library_module) -> str:
    module_file_full_path = get_module_importable_str_and_path(library_module)[1]
    return os.path.dirname(module_file_full_path)


def module_rel_path(library_top_module, library_module) -> str:
    module_file_full_path = get_module_importable_str_and_path(library_module)[1]
    top_init_file_full_path: str = get_module_importable_str_and_path(library_top_module)[1]
    library_dir_full_path: str = os.path.dirname(top_init_file_full_path)
    return get_relative_path_part(module_file_full_path, library_dir_full_path)


def module_import_str(library_top_module, library_module) -> str:
    result: str = f'{entity_name(library_top_module)}/{module_rel_path(library_top_module, library_module)}'
    result = result.replace('\\', '/')
    result = result.replace('//', '/')
    result = result.strip('/')
    result = removesuffix(result, '.py')
    result = result.strip('/')
    result = result.replace('/', '.')
    return result


def current_module_dir(depth: Optional[int] = 1) -> str:
    _, _, module_file_full_path, _ = entity_owning_module_info_and_owning_path(frame(depth + 1))
    return os.path.dirname(module_file_full_path)


def current_module_rel_path(library_top_module, depth: Optional[int] = 1) -> str:
    _, _, module_file_full_path, _ = entity_owning_module_info_and_owning_path(frame(depth + 1))
    top_init_file_full_path: str = get_module_importable_str_and_path(library_top_module)[1]
    library_dir_full_path: str = os.path.dirname(top_init_file_full_path)
    return get_relative_path_part(module_file_full_path, library_dir_full_path)


def current_module_import_str(library_top_module, depth: Optional[int] = 1) -> str:
    result: str = f'{entity_name(library_top_module)}/{current_module_rel_path(library_top_module, depth + 1)}'
    result = result.replace('\\', '/')
    result = result.replace('//', '/')
    result = result.strip('/')
    result = removesuffix(result, '.py')
    result = result.strip('/')
    result = result.replace('/', '.')
    return result



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from cengal.parallel_execution.coroutines.coro_scheduler import cs_acoro, cs_callable, current_interface, Interface
from cengal.parallel_execution.coroutines.coro_standard_services.run_coro import RunCoro
from cengal.parallel_execution.coroutines.coro_standard_services.put_coro import PutCoro
from cengal.parallel_execution.coroutines.coro_standard_services.sleep import Sleep
from cengal.parallel_execution.coroutines.coro_tools.await_coro import cs_awaitable
from cengal.parallel_execution.coroutines.coro_tools.run_in_loop import run_in_loop
import asyncio

from unittest import TestCase, main


class Test__await_coro(TestCase):
    def test__cs_awaitable(self):

        def my_coro_gi(i: Interface, a: str, b: str) -> str:
            i(Sleep, 0.1)

            return a + b

        @cs_awaitable
        @cs_callable
        def my_coro_gia(i: Interface, a: str, b: str) -> str:
            i(Sleep, 0.1)

            return a + b

        def my_coro_g(a: str, b: str) -> str:
            i: Interface = current_interface()
            i(Sleep, 0.1)

            return a + b

        @cs_awaitable
        def my_coro_ga(a: str, b: str) -> str:
            i: Interface = current_interface()
            i(Sleep, 0.1)

            return a + b

        @cs_awaitable
        async def my_coro_a(a: str, b: str) -> str:
            i: Interface = current_interface()
            await i(Sleep, 0.05)

            await asyncio.sleep(0.05)

            return a + b

        @cs_acoro
        async def my_coro_a_implicit(a: str, b: str) -> str:
            i: Interface = current_interface()
            await i(Sleep, 0.05)

            await asyncio.sleep(0.05)

            return a + b

        async def my_coro_a_explicit(i: Interface, a: str, b: str) -> str:
            await i(Sleep, 0.05)

            await asyncio.sleep(0.05)

            return a + b
        
        async def my_coro_asyncio(a: str, b: str) -> str:
            await asyncio.sleep(0.1)

            return a + b
        
        @cs_acoro
        async def amain():
            i: Interface = current_interface()

            # await

            print(await my_coro_gia('a', 'b'))
            print(await my_coro_ga('a', 'b'))
            # print(await my_coro_a('a', 'b'))
            # print(await my_coro_a_implicit('a', 'b'))
            # print(await my_coro_a_explicit(i, 'a', 'b'))
            # print(await my_coro_asyncio('a', 'b'))

            # RunCoro

            print(await i(RunCoro, my_coro_ga('a', 'b')))
            # print(await i(RunCoro, my_coro_gi, 'a', 'b'))
            # # print(await i(RunCoro, my_coro_gia, 'a', 'b'))

            # print(await i(RunCoro, my_coro_a('a', 'b')))
            # print(await i(RunCoro, my_coro_a, 'a', 'b'))
            
            # print(await i(RunCoro, my_coro_a_implicit('a', 'b')))
            # print(await i(RunCoro, my_coro_a_implicit, 'a', 'b'))
            
            # print(await i(RunCoro, my_coro_a_explicit, 'a', 'b'))
            
            # print(await i(RunCoro, my_coro_asyncio('a', 'b')))
            # print(await i(RunCoro, cs_acoro(my_coro_asyncio), 'a', 'b'))

            # PutCoro

            await i(PutCoro, my_coro_ga('a', 'b'))
            # await i(PutCoro, my_coro_gi, 'a', 'b')
            # # await i(PutCoro, my_coro_gia, 'a', 'b')

            # await i(PutCoro, my_coro_a('a', 'b'))
            # await i(PutCoro, my_coro_a, 'a', 'b')
            
            # await i(PutCoro, my_coro_a_implicit('a', 'b'))
            # await i(PutCoro, my_coro_a_implicit, 'a', 'b')
            
            # await i(PutCoro, my_coro_a_explicit, 'a', 'b')
            
            # await i(PutCoro, my_coro_asyncio('a', 'b'))
            # await i(PutCoro, cs_acoro(my_coro_asyncio), 'a', 'b')
        
        @cs_acoro
        def main():
            i: Interface = current_interface()

            # RunCoro

            print(i(RunCoro, my_coro_ga('a', 'b')))
            # print(i(RunCoro, my_coro_gi, 'a', 'b'))
            # # print(i(RunCoro, my_coro_gia, 'a', 'b'))

            # print(i(RunCoro, my_coro_a('a', 'b')))
            # print(i(RunCoro, my_coro_a, 'a', 'b'))
            
            # print(i(RunCoro, my_coro_a_implicit('a', 'b')))
            # print(i(RunCoro, my_coro_a_implicit, 'a', 'b'))
            
            # print(i(RunCoro, my_coro_a_explicit, 'a', 'b'))
            
            # print(i(RunCoro, my_coro_asyncio('a', 'b')))
            # print(i(RunCoro, cs_acoro(my_coro_asyncio), 'a', 'b'))

            # PutCoro

            i(PutCoro, my_coro_ga('a', 'b'))
            # i(PutCoro, my_coro_gi, 'a', 'b')
            # # i(PutCoro, my_coro_gia, 'a', 'b')

            # i(PutCoro, my_coro_a('a', 'b'))
            # i(PutCoro, my_coro_a, 'a', 'b')
            
            # i(PutCoro, my_coro_a_implicit('a', 'b'))
            # i(PutCoro, my_coro_a_implicit, 'a', 'b')
            
            # i(PutCoro, my_coro_a_explicit, 'a', 'b')
            
            # i(PutCoro, my_coro_asyncio('a', 'b'))
            # i(PutCoro, cs_acoro(my_coro_asyncio), 'a', 'b')
        
        run_in_loop(amain)
        run_in_loop(main)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""


__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


import justpy as jp
from justpy_components import *
from pprint import pprint
from typing import NoReturn, Union, Optional, Dict, List, Type, Callable, Sequence
from time import perf_counter
# from cengal.parallel_execution.coroutines import *
from cengal.parallel_execution.coroutines.coro_scheduler import *
from cengal.parallel_execution.coroutines.coro_standard_services.sleep import Sleep
from cengal.parallel_execution.coroutines.coro_standard_services.run_coro import RunCoro
from cengal.parallel_execution.coroutines.coro_standard_services.loop_yield import LoopYieldPriorityScheduler, CoroPriority, gly, agly
from cengal.parallel_execution.coroutines.coro_standard_services.timer_coro_runner import TimerCoroRunner
# from cengal.parallel_execution.coroutines.coro_standard_services.
# from cengal.parallel_execution.coroutines.coro_standard_services.
from cengal.parallel_execution.coroutines.coro_tools.await_coro import RunSchedulerInAsyncioLoop
from cengal.code_flow_control.args_manager import ArgsManager, EArgs
from random import randint
from time import perf_counter
from justpy_coro_helpers import coro_interfaces


Numeric = Union[int, float, complex]


# def hello_world():
#     wp = jp.WebPage()
#     jp.Hello(a=wp)
#     return wp

# jp.justpy(hello_world)

@event_handler(False)
def my_click(self, msg):
    self.text = 'I was clicked'


def event_demo():
    wp = jp.WebPage()
    d = jp.Div(text='Not clicked yet', a=wp, classes='w-48 text-xl m-2 p-1 bg-blue-500 text-white')
    d.on('click', my_click)
    return wp


theme: Theme = {
    jp.Button: 'bg-white hover:bg-gray-100 text-gray-800 font-semibold py-1 px-2 border border-gray-400 rounded-l shadow',
    jp.Label: 'bg-gray-200 text-gray-800 font-semibold py-1 px-2 border border-gray-400 rounded-r shadow',
    jp.Input: 'shadow appearance-none border rounded-l w-full py-2 px-3 text-gray-700 leading-tight focus:outline-none focus:shadow-outline h-auto',
    jp.P: 'text-gray-800 border border-gray-400 rounded-r shadow w-full h-full overflow-y-scroll',
}


class ClickableDiv(Component):
    def __init__(self, *args, **kwargs):
        self.cd = None
        if 'dtext' in kwargs:
            self.dtext = kwargs['dtext']
            del kwargs['dtext']
        super().__init__(*args, **kwargs)
    
    def init(self):
        self.html = self.theme(jp.Div, classes='-- w-auto h-auto')
        self.cd = self.theme(jp.Div, a=self.html, text='Not clicked yet', classes='w-48 h-auto text-xl mt-2 ml-2 mr-2 p-1 bg-blue-500 text-white rounded-t')
        self.cd.on('click', self.on_click)
        self.theme(jp.Div, a=self.html, text=self.dtext, classes='w-48 h-auto text-xl mb-2 ml-2 mr-2 p-1 bg-gray-500 text-white rounded-b')
    
    @event_handler()
    def on_click(self, item, msg):
        self.cd.text =  f'I was clicked: {str(item)}'


class AClickableDiv(Component):
    # def __init__(self, parent: Optional[Container], coro_scheduler: CoroSchedulerType, theme: Optional[ArgsManager]=None, dtext: Optional[str]=None):
    def __init__(self, parent: Optional[Container], await_coro_fast: Callable, await_task_fast: Callable, theme: Optional[ArgsManager]=None, dtext: Optional[str]=None):
        # self.cd = None
        # self.cs = coro_scheduler
        # # self.acf = ArgsManager(
        # #     EArgs(jp.asyncio.get_event_loop(), self.cs, CoroType.auto),
        # # )
        # self.await_coro_fast = ArgsManager(
        #     EArgs(jp.asyncio.get_event_loop(), self.cs, CoroType.auto),
        # ).callable(await_coro_fast)
        self.await_coro_fast: Callable = await_coro_fast
        self.await_task_fast: Callable = await_task_fast
        self.dtext = dtext
        self.click_counter = 0
        super().__init__(parent, theme)

    def init(self):
        self.html = self.theme(jp.Div, classes='-- w-auto h-auto')
        self.cd = self.theme(jp.Div, a=self.html, text='Click me!', classes='select-none w-48 h-auto text-xl mt-2 ml-2 mr-2 p-1 bg-blue-500 text-white rounded-t')
        # self.cd.on('click', self.on_click)
        # self.cd.on('click', AsyncEventHandlerRunner(self.on_click_ahandler))
        # self.cd.on('click', async_handle_runner(self.on_click_ahandler))
        self.cd.on('click', self.on_click_ahandler)
        self.theme(jp.Div, a=self.html, text=self.dtext, classes='w-48 h-auto text-xl mb-2 ml-2 mr-2 p-1 bg-gray-500 text-white rounded-b')
    
    @event_handler()
    async def on_click(self, item, msg):
        # print(f'{perf_counter()} > 0 - {item}')
        # self.cd.text = f'Waiting...'
        # print(f'{perf_counter()} > 1 - {item}')
        # await msg.page.update()
        # print(f'{perf_counter()} > 2 - {item}')
        # await jp.asyncio.sleep(5)
        # print(f'{perf_counter()} > 3 - {item}')
        # self.cd.text =  f'Sum: {20 + self.click_counter}'
        # print(f'{perf_counter()} > 4 - {item}')
        # await msg.page.update()
        # print(f'{perf_counter()} > 5 - {item}')
        # self.click_counter += 1
        # print(f'{perf_counter()} > 6 - {item}')
        # # coro_future = await_coro_fast(jp.asyncio.get_event_loop(), self.cs, CoroType.greenlet, AClickableDiv.on_click_coro, self)
        # # result = await coro_future
        # # self.cd.text =  f'Sum: {result}'
        jp.run_task(self.on_click_ahandler(item, msg))
    
    @staticmethod
    def add_coro(interface, a, b, item):
        print(f'{perf_counter()} > 4 - {item}')
        result = a + b
        print(f'{perf_counter()} > 5 - {item}')
        interface(Sleep, 2)
        print(f'{perf_counter()} > 6 - {item}')
        return result
    
    @staticmethod
    def on_click_coro(interface: Interface, self, item):
        print(f'{perf_counter()} > 2 - {item}')
        # interface(Sleep, 0.5)
        print(f'{perf_counter()} > 3 - {item}')
        cc = self.click_counter
        self.click_counter += 1
        result = interface(RunCoro, CoroType.auto, AClickableDiv.add_coro, 10, cc, item)
        print(f'{perf_counter()} > 7 - {item}')
        print(f'{perf_counter()} > 8 - {item}')
        return result
    
    @staticmethod
    async def on_click_handler(interface: Interface, self, item, msg):
        print(f'{perf_counter()} > 1 - {item}')
        result = await interface(RunCoro, CoroType.auto, AClickableDiv.on_click_coro, self, item)
        print(f'{perf_counter()} > 9 - {item}')
        self.cd.text =  f'_PreSum: {result}'
        print(f'{perf_counter()} > 10 - {item}')
        return result
    
    @staticmethod
    def on_click_handler_green(interface: Interface, self, item, msg):
        print(f'{perf_counter()} > 1 - {item}')
        result = interface(RunCoro, CoroType.auto, AClickableDiv.on_click_coro, self, item)
        print(f'{perf_counter()} > 9 - {item}')
        self.cd.text =  f'_PreSum: {result}'
        print(f'{perf_counter()} > 10 - {item}')
        return result

    @event_handler()
    async def on_click_ahandler(self, item, msg):
        print(f'{perf_counter()} > 0 - {item}')
        self.cd.text = 'Clicked...'
        await msg.page.update()
        # self.cs.put_coro(AClickableDiv.on_click_handler, self, item, msg)
        # result = await await_coro_fast(jp.asyncio.get_event_loop(), self.cs, CoroType.auto, AClickableDiv.on_click_handler, self, item, msg)
        # result = await await_coro_fast(jp.asyncio.get_event_loop(), self.cs, CoroType.auto, AClickableDiv.on_click_handler_green, self, item, msg)
        # result = await self.acf(await_coro_fast(AClickableDiv.on_click_handler_green, self, item, msg))
        # result = await self.await_coro_fast(AClickableDiv.on_click_handler, self, item, msg)
        # result = await self.await_coro_fast(AClickableDiv.on_click_handler_green, self, item, msg)
        def handler(interface: Interface, self, item, msg):
            print(f'{perf_counter()} > 1 - {item}')
            result = interface(RunCoro, CoroType.auto, AClickableDiv.on_click_coro, self, item)
            print(f'{perf_counter()} > 9 - {item}')
            self.cd.text =  f'_PreSum: {result}'
            print(f'{perf_counter()} > 10 - {item}')
            return result
        result = await self.await_coro_fast(handler, self, item, msg)
        print(f'{perf_counter()} > 11 - {item}')
        await msg.page.update()
        await jp.asyncio.sleep(2)
        self.cd.text = f'Sum: {result}'
        await msg.page.update()
        print(f'{perf_counter()} > 12 - {item}')


class Screen(Component):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
    
    def init(self):
        self.html = self.container = self.theme(jp.Div, classes='-- h-screen flex')


class Pannels(Component):
    def __init__(self, *args, **kwargs):
        if 'pan_l_classes' in kwargs:
            self.pan_l_classes = kwargs['pan_l_classes']
            del kwargs['pan_l_classes']
        super().__init__(*args, **kwargs)
    
    def init(self):
        self.html = self.theme(jp.Div, classes='-- h-full flex')
        self.containers['pan_l'] = self.theme(jp.Div, a=self.html, classes='-- h-auto overflow-y-auto '+self.pan_l_classes)
        pan_r_holder = self.theme(jp.Div, a=self.html, classes='-- flex-1 flex overflow-hidden')
        self.container = self.containers['pan_r'] = self.theme(jp.Div, a=pan_r_holder, classes='-- flex-1 overflow-y-auto')

# GLOBAL_LOOPER = None

class Looper:
    def __init__(self):
        self.external_loop = jp.asyncio.get_event_loop()
        self.cs = CoroScheduler()
        self.configure_coro_scheduler()
        self.iter_num = 0
        self.delta_num = 0
        self.register_looper()
        self.rs = RunSchedulerInAsyncioLoop(self.cs, 0.020, self.external_loop)
        self.rs.register()
    
    def register_looper(self):
        self.external_loop.call_soon(self.looper, self.external_loop, 0, 0.0, 0)
    
    def looper(self, loop, iter_num: int, last_rendering_time: float, last_renderred_num: int):
        current_time = perf_counter()
        if 1.0 <= (current_time - last_rendering_time):
            last_rendering_time = current_time
            delta_num = iter_num - last_renderred_num
            last_renderred_num = iter_num
            self.iter_num = iter_num
            self.delta_num = delta_num
            print(f'<><><> Looper <><><> iteration id: {iter_num}; iteration delta: {delta_num} <><><>')
        loop.call_soon(self.looper, loop, iter_num + 1,last_rendering_time, last_renderred_num)
    
    def make_sum(self, interface: Interface, a: Numeric, b: Numeric) -> Numeric:
        print('Greenlet')
        interface(Sleep, 1)
        ly = gly(CoroPriority.normal)
        result = 0
        for i in range(randint(1, 100000)):
            result += a + b + i
            ly()
        return result
    
    async def amake_sum(self, interface: Interface, a: Numeric, b: Numeric) -> Numeric:
        print('Awaitable')
        await interface(Sleep, 1)
        ly = await agly(CoroPriority.normal)
        result = 0
        for i in range(randint(1, 100000)):
            result += a + b + i
            await ly()
        return result

    def main_coro(self, interface: Interface):
        last_cs = interface._loop.context_switches
        while True:
            # res = interface(RunCoro, CoroType.greenlet, self.make_sum, self.iter_num, self.delta_num)
            # res = interface(RunCoro, CoroType.awaitable, self.amake_sum, self.iter_num, self.delta_num)
            coro_list = (self.make_sum, self.amake_sum)
            chosen_coro = coro_list[randint(0, len(coro_list) - 1)]
            res = interface(RunCoro, CoroType.auto, chosen_coro, self.iter_num, self.delta_num)
            print(f'<><><> CORO <><><> iteration id: {self.iter_num}; iteration delta: {self.delta_num}; sum: {res}; delta context switches: {interface._loop.context_switches - last_cs} <><><>')
            last_cs = interface._loop.context_switches
            interface(Sleep, 2)
            for i in range(randint(1, 25000)):
                result = self.iter_num + self.delta_num
                gly(CoroPriority.normal)()

    def configure_coro_scheduler(self):
        self.cs.register_service(RunCoro)
        self.cs.register_service(LoopYieldPriorityScheduler)
        self.cs.register_service(TimerCoroRunner)
        self.cs.register_service(Sleep)
        self.cs.register_service(RunCoro)
        self.cs.put_coro(self.main_coro)


# def components_demo():
#     Looper()
    
#     # pprint(globals())
    
#     tm = ThemeManager(theme)
#     wp = jp.WebPage()
    
#     screen = tm(Screen, wp)
    
#     tm(ClickableDiv, screen, dtext='0')
#     main_pannel = tm(Pannels, screen, pan_l_classes='w-auto')
#     tm(ClickableDiv, main_pannel, dtext='1')
#     tm(ClickableDiv, main_pannel.containers['pan_l'], dtext='2.0')
#     tm(ClickableDiv, main_pannel.containers['pan_l'], dtext='2.1')
#     tm(ClickableDiv, main_pannel.containers['pan_l'], dtext='2.2')
#     tm(ClickableDiv, main_pannel.containers['pan_l'], dtext='2.3')
#     tm(ClickableDiv, main_pannel.containers['pan_l'], dtext='2.4')
#     tm(ClickableDiv, main_pannel.containers['pan_l'], dtext='2.5')
#     tm(ClickableDiv, main_pannel.containers['pan_l'], dtext='2.6')
    
#     limiter = jp.Div(a=main_pannel.container, classes="h-24")
#     # limiter = jp.Div(a=main_pannel.container, classes="h-auto")
    
#     sub_pannel = tm(Pannels, limiter, pan_l_classes='w-auto')
#     tm(ClickableDiv, sub_pannel, dtext='3')
#     tm(ClickableDiv, sub_pannel, dtext='4')
#     tm(ClickableDiv, sub_pannel, dtext='5')
#     tm(ClickableDiv, sub_pannel, dtext='6')
#     tm(ClickableDiv, sub_pannel, dtext='7')
#     tm(ClickableDiv, sub_pannel, dtext='8')
#     tm(ClickableDiv, sub_pannel, dtext='9')
#     tm(ClickableDiv, sub_pannel, dtext='10')
#     tm(ClickableDiv, sub_pannel, dtext='11')
#     tm(ClickableDiv, sub_pannel.containers['pan_l'], dtext='12')
#     tm(ClickableDiv, sub_pannel.containers['pan_l'], dtext='13')
#     tm(ClickableDiv, sub_pannel.containers['pan_l'], dtext='14')

#     tm(ClickableDiv, main_pannel, dtext='15')
#     # f1 = tm(ClickableDiv, wp, dtext='16')
#     # tm(ClickableDiv, f1, dtext='17')
#     return wp

# jp.justpy(components_demo)

async def acomponents_demo():
    # global GLOBAL_LOOPER
    # GLOBAL_LOOPER = Looper()
    
    # pprint(globals())

    external_loop = jp.asyncio.get_event_loop()
    cs = CoroScheduler()
    cs.register_service(RunCoro)
    cs.register_service(LoopYieldPriorityScheduler)
    cs.register_service(TimerCoroRunner)
    cs.register_service(Sleep)
    cs.register_service(RunCoro)
    rs = RunSchedulerInAsyncioLoop(cs, 0.020, external_loop)
    rs.register()
    
    # lp = Looper()
    # cs = lp.cs
    
    tm = ArgsManager(
        ThemeArgsManager(theme)
    )
    am = ArgsManager(
        ThemeArgsManager(theme),
        coro_interfaces(cs)
    )
    
    wp = jp.WebPage()
    
    screen = tm(Screen, wp)
    
    am(AClickableDiv, screen, dtext='0')
    main_pannel = tm(Pannels, screen, pan_l_classes='w-auto')
    am(AClickableDiv, main_pannel, dtext='1')
    am(AClickableDiv, main_pannel.containers['pan_l'], dtext='2.0')
    am(AClickableDiv, main_pannel.containers['pan_l'], dtext='2.1')
    am(AClickableDiv, main_pannel.containers['pan_l'], dtext='2.2')
    am(AClickableDiv, main_pannel.containers['pan_l'], dtext='2.3')
    am(AClickableDiv, main_pannel.containers['pan_l'], dtext='2.4')
    am(AClickableDiv, main_pannel.containers['pan_l'], dtext='2.5')
    am(AClickableDiv, main_pannel.containers['pan_l'], dtext='2.6')
    
    limiter = jp.Div(a=main_pannel.container, classes="h-24")
    # limiter = jp.Div(a=main_pannel.container, classes="h-auto")
    
    sub_pannel = tm(Pannels, limiter, pan_l_classes='w-auto')
    am(AClickableDiv, sub_pannel, dtext='3')
    am(AClickableDiv, sub_pannel, dtext='4')
    am(AClickableDiv, sub_pannel, dtext='5')
    am(AClickableDiv, sub_pannel, dtext='6')
    am(AClickableDiv, sub_pannel, dtext='7')
    am(AClickableDiv, sub_pannel, dtext='8')
    am(AClickableDiv, sub_pannel, dtext='9')
    am(AClickableDiv, sub_pannel, dtext='10')
    am(AClickableDiv, sub_pannel, dtext='11')
    am(AClickableDiv, sub_pannel.containers['pan_l'], dtext='12')
    am(AClickableDiv, sub_pannel.containers['pan_l'], dtext='13')
    am(AClickableDiv, sub_pannel.containers['pan_l'], dtext='14')

    am(AClickableDiv, main_pannel, dtext='15')
    # f1 = am(AClickableDiv, wp, dtext='16')
    # am(AClickableDiv, f1, dtext='17')
    return wp

jp.justpy(acomponents_demo)


# def pannel_demo():
#     tm = ThemeManager(theme)
#     wp = jp.WebPage()
#     screen = tm(Screen, wp)
#     tm(ClickableDiv, screen)
#     main_pannel = tm(Pannels, screen)
#     tm(ClickableDiv, main_pannel)
#     tm(ClickableDiv, main_pannel.containers['pan_l'])
#     return wp

# jp.justpy(pannel_demo)





#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = [
    'SetWindowCompositionAttribute',
    'GetWindowPlacement',
    'MonitorFromWindow',
    'GetMonitorInfoW',
    'ShowWindow',
    'CreateWindowExW',
    'LoadCursorW',
    'RegisterClassExW',
    'GetWindowLongPtrW',
    'SetWindowLongPtrW_LongPtr',
    'SetWindowLongPtrW_WndProc',
    'CallWindowProc',
    'SetWindowPos',
    'ShowWindow',
    'DefWindowProcW',
    'DestroyWindow',
    'PostQuitMessage',
    'GetSystemMetrics',
    'GetWindowRect',
    'GetMessageW',
    'TranslateMessage',
    'DispatchMessageW',
    'MessageBoxW',
    'CreateSolidBrush',
    'DwmSetWindowAttribute',
    'GetCurrentThreadId',
    'GetDpiForWindow',
    'UpdateWindow',
    'SHGetKnownFolderPath',
    'CoTaskMemFree',
]


import ctypes
from ctypes import HRESULT, byref, pointer, wintypes, windll, WINFUNCTYPE
import win32gui, win32con, winerror

from ctypes import sizeof, byref, pointer, POINTER, c_int, c_long, c_longlong, c_ulonglong, c_wchar_p
from ctypes.wintypes import ATOM, COLORREF, HANDLE, HBRUSH, HINSTANCE, HMENU, HWND, BOOL, HMONITOR, DWORD, LPCVOID, LPCWSTR, LPMSG, LPRECT, LPVOID, LONG, MSG, RGB, UINT, WPARAM, LPARAM, HHOOK
from win32con import IDC_ARROW, CS_HREDRAW, CS_VREDRAW, GWL_STYLE, GWL_WNDPROC, SWP_FRAMECHANGED, SWP_NOMOVE, SWP_NOSIZE, SWP_NOZORDER, SW_SHOW, SM_CXFRAME, SM_CYFRAME


from cengal.ctypes_tools.types import *
from cengal.ctypes_tools.libraries import user32, gdi32, dwmapi, kernel32, shell32, ole32
from cengal.ctypes_tools.tools import cwfunc_def


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


SetWindowCompositionAttribute = cwfunc_def(user32, 'SetWindowCompositionAttribute', (HWND, POINTER(WINCOMPATTRDATA)), BOOL)
GetWindowPlacement = cwfunc_def(user32, 'GetWindowPlacement', (HWND, POINTER(WINDOWPLACEMENT)), BOOL)
MonitorFromWindow = cwfunc_def(user32, 'MonitorFromWindow', (HWND, DWORD), HMONITOR)
GetMonitorInfoW = cwfunc_def(user32, 'GetMonitorInfoW', (HMONITOR, POINTER(MONITORINFO)), BOOL)
ShowWindow = cwfunc_def(user32, 'ShowWindow', (HWND, c_int), BOOL)
CreateWindowExW = cwfunc_def(user32, 'CreateWindowExW', (DWORD, LPCWSTR, LPCWSTR, DWORD, c_int, c_int, c_int, c_int, HWND, HMENU, HINSTANCE, LPVOID), HWND)
LoadCursorW = cwfunc_def(user32, 'LoadCursorW', (HINSTANCE, LPCWSTR), HCURSOR)
RegisterClassExW = cwfunc_def(user32, 'RegisterClassExW', (POINTER(WNDCLASSEXW),), ATOM)
GetWindowLongPtrW = cwfunc_def(user32, 'GetWindowLongPtrW', (HWND, c_int), LONG_PTR)
SetWindowLongPtrW_LongPtr = cwfunc_def(user32, 'SetWindowLongPtrW', (HWND, c_int, LONG_PTR), LONG_PTR)
SetWindowLongPtrW_WndProc = cwfunc_def(user32, 'SetWindowLongPtrW', (HWND, c_int, WNDPROC), WNDPROC)
CallWindowProc = cwfunc_def(user32, 'CallWindowProcW', (WNDPROC, HWND, UINT, WPARAM, LPARAM), LRESULT)
SetWindowPos = cwfunc_def(user32, 'SetWindowPos', (HWND, HWND, c_int, c_int, c_int, c_int, UINT), BOOL)
ShowWindow = cwfunc_def(user32, 'ShowWindow', (HWND, c_int), BOOL)
DefWindowProcW = cwfunc_def(user32, 'DefWindowProcW', (HWND, UINT, WPARAM, LPARAM), LRESULT)
DestroyWindow = cwfunc_def(user32, 'DestroyWindow', (HWND,), BOOL)
PostQuitMessage = cwfunc_def(user32, 'PostQuitMessage', (c_int,), None)
GetSystemMetrics = cwfunc_def(user32, 'GetSystemMetrics', (c_int,), c_int)
GetWindowRect = cwfunc_def(user32, 'GetWindowRect', (HWND, LPRECT), BOOL)
GetMessageW = cwfunc_def(user32, 'GetMessageW', (LPMSG, HWND, UINT, UINT), BOOL)
TranslateMessage = cwfunc_def(user32, 'TranslateMessage', (POINTER(MSG),), BOOL)
DispatchMessageW = cwfunc_def(user32, 'DispatchMessageW', (POINTER(MSG),), LRESULT)
MessageBoxW = cwfunc_def(user32, 'MessageBoxW', (HWND, LPCWSTR, LPCWSTR, UINT), c_int)
CreateSolidBrush = cwfunc_def(gdi32, 'CreateSolidBrush', (COLORREF,), HBRUSH)
DwmSetWindowAttribute = cwfunc_def(dwmapi, 'DwmSetWindowAttribute', (HWND, DWORD, LPCVOID, DWORD), HRESULT)
# SetWindowsHookEx = cwfunc_def(user32, SetWindowsHookExW, (c_int, WINFUNCTYPE, HINSTANCE, DWORD), HHOOK)
GetCurrentThreadId = cwfunc_def(kernel32, 'GetCurrentThreadId', tuple(), DWORD)
GetDpiForWindow = cwfunc_def(user32, 'GetDpiForWindow', (HWND,), UINT)
UpdateWindow = cwfunc_def(user32, 'UpdateWindow', (HWND,), BOOL)
SHGetKnownFolderPath = cwfunc_def(shell32, 'SHGetKnownFolderPath', (POINTER(GUID), DWORD, HANDLE, POINTER(c_wchar_p)), HRESULT)
CoTaskMemFree = cwfunc_def(ole32, 'CoTaskMemFree', (LPVOID,), None)
